#!/usr/bin/env python3

"""
This file contains a number of classes and free functions dealing with
bulk data access. There is a similar file impl.meta that deals with
metadata. Nothing in this file should be directly visible to users of
the public API.

    impl.bulk.ZgyInternalBulk:

        * Internal methods to read and write bulk.
        * Needs access to much of the meta data information.
          Currently this is handled by giving this class references to
          those file headers it needs, e.g. InfoHeader, BrickLUP, etc.
          TODO-Low: Refactor, would it be cleaner to instead provide a
          reference to the ZgyInternalMeta instance? This gives access to
          more metadata than the class actually needs. But would provide
          better isolation nevertheless.
        * Refactoring notes: There are a couple of sections in this
          class that ought to have been separated out. As it is now
          the class does too much.
            - Write support
            - Lod generation
            - Histogram generation

"""

##@package openzgy.impl.bulk
# \brief Reading and writing bulk data.

import numpy as np
import struct
import os
import math
import sys

from ..impl import enum as impl_enum
from ..impl import file as impl_file
from ..exception import *
from ..impl.lodalgo import decimate, DecimationType
from ..impl.stats import StatisticData
from ..impl.compress import CompressFactoryImpl

# Changing this should be done ONLY FOR TESTING.
# Use a different value for 'outside survey area' than the default.
_padding_fill_value = None

def dprint(*args, **kwargs):
    return None # comment out to get debug logging
    #print(*args, **kwargs)

class ErrorsWillCorruptFile:
    """
    Duplicated between impl.bulk and impl.meta. Maybe fix sometime.
    Technically the two differ because they set different flags.
    In ZgyInternalBulk and ZgyInternalMeta respectively. But that
    distinction isn't easy to see in Python.

    Start a critical section where any exception means that the
    owner class should be permanently flagged with _is_bad = True.
    Typically this is used to prevent secondary errors after a
    write failure that has most likely corrupted the entire file.
    The exception itself will not be caught.

    The _is_bad flag normally means that any further attempts
    to access this class, at least for writing, will raise a
    ZgyCorruptedFile exception. Regardless of what the exception
    was that caused the flag to be set.
    """
    def __init__(self, parent): self._parent = parent
    def __enter__(self): return None
    def __exit__(self, type, value, traceback):
        if type:
            #print("Bulk: Exit critical section", str(value))
            self._parent._is_bad = True

class ScalarBuffer:
    """
    Represents an ndarray where all elements have the same value.
    This is basically a scalar but additionally has a .shape
    attribute telling how large an array it represents.
    """
    def __init__(self, shape, value, dtype):
        self._shape = tuple(shape)
        self._dtype = np.dtype(dtype)
        self._value = self._dtype.type(value)
    @property
    def shape(self): return self._shape
    @property
    def dtype(self): return self._dtype
    @property
    def value(self): return self._value
    @property
    def itemsize(self): return self._dtype.itemsize
    def __len__(self): return 0
    def __str__(self): return str(self._value)
    def __repr__(self): return "numpy.{0}({1})".format(self._dtype.name, self._value)
    def inflate(self): return np.full(self._shape, self._value, dtype=self._dtype)

class ZgyInternalBulk:
    """
    Read or write bulk data. The meta data needs to have been read
    already. The user-callable API will forward its read requests here.
    """
    def __init__(self, f, metadata, *, compressed = False):
        self._is_bad = False
        self._file = f
        # Note, ZgyInternalBulk._metadata and ZgyMeta._meta will point to
        # the same instance of ZgyInternalMeta. The members deliberately
        # have different names; I believe this will to reduce confusion.
        self._metadata = metadata
        self._compressed = compressed # If true, do not align bricks.
        # Technically only needed if self._metadata._fh._version == 1
        # but do it unconditionally to get better test coverage.
        self._unscramble_init()
        # Keep track of value range as LOD 0 samples are written.
        # If samples are overwritten the range might be too wide.
        # This range will be used when computing the histogram.
        # The range is always user values, not storage. This migh
        # be of academic interest only, as the information is
        # currently ignored when the data on the file is integral.
        self._sample_min = np.inf
        self._sample_max = -np.inf
        # TODO-Low some way for application code to configure this.
        # For now don't check for cloud/local because uncompressed
        # local files are reasonable to have as Always, especially
        # since it is so backward to change.
        #self._update_mode = impl_enum.UpdateMode.Always if not compressed and not f.xx_iscloud else impl_enum.UpdateMode.Constant
        self._update_mode = impl_enum.UpdateMode.Always if not compressed else impl_enum.UpdateMode.Constant

    @staticmethod
    def _argType(parameter, accept, name = "parameter"):
        """
        Check that the parameter has one of the acceped types.
        If None is allowed, include type(None) in the accepted list.
        Raise an exception on error.
        The main reason I am using this function is that I changed the
        convention for passing data buffers around. This used to be
        either a np.ndarray or a scalar, with the latter often requiring
        a "size" parameter as well. The new convention is to pass either
        np.ndarray of ScalarBuffer. Given Python's weak typing this
        might easily introduce some bugs.
        """
        if not isinstance(parameter, accept):
            if not isinstance(accept, tuple): accept = (accept,)
            oknames = tuple([t.__name__ for t in accept])
            myname = type(parameter).__module__ + "." + type(parameter).__name__
            err = "{0} must be in {1} but was {2}".format(name, oknames, myname)
            raise TypeError(err)

    def _checkCompatibleBufferTypes(self, user_dtype, is_storage):
        """
        Several methods now have both an explicit is_storage argument
        and a return-buffer value type that can be used to infer whether
        the buffer or scalar needs to be converted from a float 'user'
        value to an integer 'storage' values. The is_storage parameter
        has precedence but to help trap bugs the value type must match.
        """
        file_dtype = impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype)
        if is_storage and user_dtype != file_dtype:
            raise ZgyInternalError("Wrong data type " + str(data.dtype) +
                                   " for a file with " + str(file_dtype))
        elif not is_storage and user_dtype != np.float32:
            raise ZgyInternalError("Wrong data type " + str(data.dtype) +
                                   ", when is_storage=False it should be float")

    @staticmethod
    def _slice(beg, end):
        """For debugging output only."""
        return "[{0}:{3},{1}:{4},{2}:{5}]".format(*beg, *end)
        #return "[{0}:{3},{1}:{4},{2}:{5}] (size {6},{7},{8})".format(*beg, *end, *(end-beg))

    @classmethod
    def _getOverlap(cls, src_beg, src_count, dst_beg, dst_count, survey_beg, survey_count):
        """
        Determine the overlap between two or three N-dimensional regions.
        Typically used to compare what the user wanted with the content
        of one particular 3d data block, and possibly also with the
        declared total area of the survey. With the intent of copying
        only the overlapping area. Return the offsets relative to both
        the source and the target region.
        """
        src_beg = np.array(src_beg, dtype=np.int64)
        src_end = src_beg + np.array(src_count, dtype=np.int64)
        dst_beg = np.array(dst_beg, dtype=np.int64)
        dst_end = dst_beg + np.array(dst_count, dtype=np.int64)
        # Get the overlap, relative to cube origin, between the brick
        # we just read (src) and the area that the user requested (dst).
        overlap_beg = np.maximum(src_beg, dst_beg)
        overlap_end = np.minimum(src_end, dst_end)
        # Optionally clip to the survey size as well.
        if survey_beg is not None and survey_count is not None:
            survey_beg = np.array(survey_beg, dtype=np.int64)
            survey_end = survey_beg + np.array(survey_count, dtype=np.int64)
            overlap_beg = np.maximum(overlap_beg, survey_beg)
            overlap_end = np.minimum(overlap_end, survey_end)
        #print("// src relative to cube:", cls._slice(src_beg, src_end))
        #print("// dst relative to cube:", cls._slice(dst_beg, dst_end))
        #print("// overlap to be copied:", cls._slice(overlap_beg, overlap_end))
        # Now convert from offsets relative to the cube orign
        # to offsets into the two buffers.
        local_src_beg = overlap_beg - src_beg
        local_src_end = overlap_end - src_beg
        local_dst_beg = overlap_beg - dst_beg
        local_dst_end = overlap_end - dst_beg
        #print("// src relative to buffer:", cls._slice(local_src_beg, local_src_end))
        #print("// dst relative to buffer:", cls._slice(local_dst_beg, local_dst_end))
        return local_src_beg, local_src_end, local_dst_beg, local_dst_end

    @classmethod
    def _partialCopy(cls, src, src_beg, src_count, dst, dst_beg, dst_count, survey_beg, survey_count, *, verbose = None):
        """
        Copy bulk data from src to dst, but only in the overlapping region.
        The src buffer will typically be a single brick read from file.
        Src can also be a scalar, in which case the overlapping region
        will have all its samples set to this particular value.
        The method will typically be called several times, once for each
        brick that was read. Both src and dest need to be either numpy
        arrays or (in the case of src) a scalar. Or a ScalarBuffer.
        """
        if isinstance(src, (np.ndarray, ScalarBuffer)): assert tuple(src.shape) == tuple(src_count)
        if isinstance(dst, np.ndarray): assert tuple(dst.shape) == tuple(dst_count)
        dst_count = dst.shape
        (local_src_beg, local_src_end,
         local_dst_beg, local_dst_end) = cls._getOverlap(
             src_beg, src_count, dst_beg, dst_count, survey_beg, survey_count)
        if np.all(local_src_end > local_src_beg):
            if isinstance(src, np.ndarray):
                if verbose:
                    verbose("set result",
                          cls._slice(local_dst_beg, local_dst_end), "= brick",
                          cls._slice(local_src_beg, local_src_end))
                dst[local_dst_beg[0]:local_dst_end[0],
                    local_dst_beg[1]:local_dst_end[1],
                    local_dst_beg[2]:local_dst_end[2]] = (
                src[local_src_beg[0]:local_src_end[0],
                    local_src_beg[1]:local_src_end[1],
                    local_src_beg[2]:local_src_end[2]])
            else:
                if verbose:
                    verbose("set result",
                          cls._slice(local_dst_beg, local_dst_end),
                          "= constant", src)
                value = src.value if isinstance(src, ScalarBuffer) else src
                dst[local_dst_beg[0]:local_dst_end[0],
                    local_dst_beg[1]:local_dst_end[1],
                    local_dst_beg[2]:local_dst_end[2]] = value
        else:
            if verbose:
                verbose("no ovelap with this brick")

    @staticmethod
    def _scaleFactorsStorageToFloat(codingrange, file_dtype):
        """
        Get the linear transform y = a*x + b for converting from
        storage values to actual floating point values.
        The smallest value in storage (e.g. -128) should map to codingrange[0]
        and the largest value (e.g. +127) should map to codingrange[1].
        If file_dtype is a floating point type there is never any conversion.
        """
        if not np.issubdtype(file_dtype, np.integer):
            (a, b) = (1, 0)
        else:
            iinfo = np.iinfo(file_dtype)
            a = (codingrange[1] - codingrange[0]) / (iinfo.max - iinfo.min)
            b = codingrange[0] - a * iinfo.min
        return (a, b)

    def _scaleDataFactorsStorageToFloat(self):
        ih = self._metadata._ih
        file_dtype = impl_enum._map_DataTypeToNumpyType(ih._datatype)
        return self._scaleFactorsStorageToFloat(ih._safe_codingrange, file_dtype)

    @classmethod
    def _scaleToFloat(cls, data, codingrange, file_dtype):
        """
        Scale integral data from storage according to the coding range.
        Input can be either a scalar or a numpy array.
        The operation is a no-op if file_dtype is float.
        The data must be known to be in storage domain, we don't try
        to guess based on its valuetype.
        """
        if data is None: return None
        # The compression module might not like the following assert
        # because for compressed integral data the decompressor might
        # convert from integral to float but leave the scaling for
        # later. Currently not a problem.
        if isinstance(data, (ScalarBuffer, np.ndarray)): assert file_dtype == data.dtype
        if not np.issubdtype(file_dtype, np.integer): return data
        (a, b) = cls._scaleFactorsStorageToFloat(codingrange, file_dtype)
        if isinstance(data, np.ndarray):
            # Numerical accuracy notes:
            #
            # To match the (arguably broken) old ZGY implementation
            # do all the computations, including the codingrange to
            # (slope, intercept) calculation, using single precision
            # float. The result isn't seriously wrong but values
            # close to zero may see a noticeable shift.
            #
            # To get the result as accurate as possible, convert the
            # entire array to float64, then apply the conversion,
            # then convert back to float32. The compromise currently
            # used is to accurately compute slope, intercept but do
            # the scaling on float32 values. The difference compared
            # to the most accurate case is hardly noticeable.
            #
            # See also InfoHeaderAccess::storagetofloat() in OpenZGY/C++.
            #
            data = data.astype(np.float32)
            # In-place operators to avoid copying arrays again.
            data *= a
            data += b
            return data
        elif isinstance(data, ScalarBuffer):
            return ScalarBuffer(data.shape, a * float(data.value) + b, np.float32)
        else:
            return np.float32(a * float(data) + b)

    def _scaleDataToFloat(self, data):
        """
        Convert and/or scale the storage value type to user-supplied data.
        Unlike _scaleToFloat() this is an instance method so it already
        knows the storage type. The data must be known to be in storage
        domain, we don't try to guess based on its valuetype.
        """
        file_dtype = impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype)
        return self._scaleToFloat(data, self._metadata._ih._safe_codingrange, file_dtype)

    @staticmethod
    def _clipAndCast(data, dtype):
        """
        Cast to the specified numpy type, clipping any values outside
        the valid range to the closest edge. Works both for scalars
        and numpy arrays. Can be made to support lists and tuples also,
        but this is a case of yagni.
        If the data is a scalar then the result will be of the reqested
        numpy type and not the generic python 'int'.
        """
        cliprange = (np.iinfo(dtype).min, np.iinfo(dtype).max)
        if isinstance(data, np.ndarray):
            np.clip(data, cliprange[0], cliprange[1], out=data) # in-place clip
            # Can I save a few cycles by passing dtype to np.rint and
            # dropping the call to astype? Not a big deal and the
            # documentation is a bit unclear.
            data = np.rint(data).astype(dtype)
        elif isinstance(data, ScalarBuffer):
            value = np.rint(max(cliprange[0], min(cliprange[1], data.value)))
            data = ScalarBuffer(data.shape, value, dtype)
        else:
            data = np.rint(max(cliprange[0], min(cliprange[1], data)))
            data = dtype(data)
        return data

    @classmethod
    def _scaleToStorage(cls, data, codingrange, file_dtype):
        """
        Scale floating point data to integral storage values according
        to the coding range. Input can be either a scalar or a numpy array.
        Or the new ScalarBuffer type.
        Input values that cannot be encoded will be clipped to the closest
        value that can be represented.
        The operation is a no-op if file_dtype is float and not an integral type.
        The input data must be np.float32 or an array of same; this is to
        avoid accidentally passing an input that is already in storage.
        """
        # Test: unittests.test_scaleToStorage()
        vt = data.dtype if isinstance(data, (np.ndarray, ScalarBuffer)) else type(data)
        if vt != np.float32:
            raise ZgyInternalError("Input to _scaleToStorage must be np.float32")
        if not np.issubdtype(file_dtype, np.integer): return data
        (a, b) = cls._scaleFactorsStorageToFloat(codingrange, file_dtype)
        if isinstance(data, np.ndarray):
            data = data.astype(np.float32)
            # In-place operators to avoid copying arrays again.
            data -= b
            data *= (1.0/a)
        elif isinstance(data, ScalarBuffer):
            value = (float(data.value) - b) / a
            data = ScalarBuffer(data.shape, value, np.float32)
        else:
            data = (float(data) - b) / a
        return cls._clipAndCast(data, file_dtype)

    def _defaultValue(self, as_float):
        """
        Get the "real" or "storage" value to be used when samples that
        have never been written. as_float=False means the return will be
        a "storage" value as found in the file. Otherwise convert to "real".
        If the data is a scalar then the result will be of the reqested
        numpy type and not the generic python 'int'.
        TODO-Performance, compute this when the file is opened.
        """
        defaultstorage = self._scaleDataToStorage(np.float32(0))
        defaultreal = self._scaleDataToFloat(defaultstorage)
        return defaultreal if as_float else defaultstorage

    @staticmethod
    def _bricksNeeded(start, size, bricksize):
        """
        Return the list of bricks needed to cover the entire area given by
        start and size. Each entry in the list gives both the sample position
        and the brick position, each as a 3-tuple of integers.
        """
        brick0 = [start[0] // bricksize[0],
                  start[1] // bricksize[1],
                  start[2] // bricksize[2]]
        brickN = [(start[0] + size[0] - 1) // bricksize[0],
                  (start[1] + size[1] - 1) // bricksize[1],
                  (start[2] + size[2] - 1) // bricksize[2]]
        result = []
        for ii in range(brick0[0], brickN[0]+1):
            for jj in range(brick0[1], brickN[1]+1):
                for kk in range(brick0[2], brickN[2]+1):
                    result.append(((ii * bricksize[0],
                                    jj * bricksize[1],
                                    kk * bricksize[2]),
                                   (ii, jj, kk)))
        return result

    def _validatePosition(self, i, j, k, lod):
        if lod < 0 or lod >= self._metadata._ih._nlods:
            raise ZgyUserError("Requested LOD {0} is outside the valid range {1} to {2} inclusive".format(lod, 0, self._metadata._ih._nlods - 1))
        size = self._metadata._ih._lodsizes[lod]
        if (i<0 or i>=size[0] or j<0 or j>=size[1] or k<0 or k>=size[2]):
            raise ZgyUserError("Requested brick position {0} is outside the valid range {1} to {2} inclusive at lod {3}".format((i, j, k), (0, 0, 0), tuple(np.array(size) - 1), lod))

    def _getBrickLookupIndex(self, i, j, k, lod):
        self._validatePosition(i, j, k, lod)
        size = self._metadata._ih._lodsizes[lod]
        index = (self._metadata._ih._brickoffsets[lod] +
  	         i + (size[0] * j) + (size[0] * size[1] * k))
        if index < 0 or index >= self._metadata._ih._brickoffsets[-1]:
            raise ZgyInternalError("Internal error in _getBrickLookupIndex")
        return index

    def _getAlphaSizeInBytes(self):
        return int(self._metadata._ih._bricksize[0] * self._metadata._ih._bricksize[1])

    def _getBrickSizeInBytes(self):
        """
        Get the size of an uncompressed brick in bytes.
        TODO-Performance, this should be cached on file open and
        should probably be a derived attribute of self._metadata._ih.
        NOTE-Performance: np.product() might be preferable to spelling out
        the multiply and needing a temp. But it could be 100 times slower.
        """
        file_dtype = np.dtype(impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype))
        bs = self._metadata._ih._bricksize
        maxsize = bs[0] * bs[1] * bs[2] * file_dtype.itemsize
        return int(maxsize)

    def _getBegAndSize(self, lup, ix, maxsize):
        """
        Use only for bricks known to be compressed. Unless testing.
        Return both the file offset of the compressed block and its size.
        The latter is a hint that should not be fully trusted.
        For testing pass sane = False, this skips the test for too large
        bricks etc. which means I can use the tool to find holes in the
        allocated data.
        """
        # TODO-Low consider deferring _calc_lookupsize() until first needed.
        # The simple version below id NOT THREADSAFE.
        #if self._metadata._blup._lookend is None:
        #    self._metadata._blup._lookend = (
        #        self._metadata._blup._calc_lookupsize(self._lookup, eof, maxsize))
        raw_beg = lup._lookup[ix]
        raw_end = lup._lookend[ix]
        btype = (raw_beg >> 56) & 0xFF
        if btype == 0x80 or raw_beg < 2:
            return 0, 0
        beg = raw_beg & ~(0xFF << 56)
        end = raw_end
        # Need extra tests if _calc_lookupsize() was called without
        # giving  the eof and maxsize parameters.
        end = max(beg, min(end, self._file.xx_eof, beg + maxsize))
        return (beg, end - beg)

    def _getAlphaBegAndSize(self, ix, *, sane = True):
        return self._getBegAndSize(self._metadata._alup, ix, self._getAlphaSizeInBytes() if sane else (0xFF<<56))

    def _getBrickBegAndSize(self, ix, *, sane = True):
        return self._getBegAndSize(self._metadata._blup, ix, self._getBrickSizeInBytes() if sane else (0xFF<<56))

    def _getBrickFilePosition(self, i, j, k, lod):
        """
        returns (brickstatus, fileoffset, constvalue, bricksize).
        """
        ix = self._getBrickLookupIndex(i, j, k, lod)
        pos = self._metadata._blup._lookup[ix]
        btype = (pos >> 56) & 0xFF
        if pos == 0:
            return (impl_enum.BrickStatus.Missing, None, 0, 0)
        elif pos == 1:
            return (impl_enum.BrickStatus.Constant, None, 0, 0)
        elif btype == 0x80:
            # TODO-Worry probably won't work with big-endian.
            # I suspect the old C++ reader won't either,
            # but that is an academic question since we
            # only support that library for x86.
            fmt = impl_enum._map_DataTypeToStructFormatCode(self._metadata._ih._datatype)
            tmp = struct.pack("<Q", pos)
            constant = struct.unpack_from(fmt, tmp)[0]
            return (impl_enum.BrickStatus.Constant, None, constant, 0)
        elif btype == 0xC0:
            beg, size = self._getBrickBegAndSize(ix)
            return (impl_enum.BrickStatus.Compressed, beg, None, size)
        elif pos & (1<<63):
            raise ZgyFormatError("Unknown brick type " + str(btype))
        else:
            return (impl_enum.BrickStatus.Normal, pos, None, self._getBrickSizeInBytes())

    def _setBrickFilePosition(self, i, j, k, lod, brickstatus, content, bricksize):
        ix = self._getBrickLookupIndex(i, j, k, lod)
        if brickstatus == impl_enum.BrickStatus.Missing:
            self._metadata._blup._lookup[ix] = 0
            self._metadata._blup._lookend[ix] = 0
        elif brickstatus == impl_enum.BrickStatus.Constant:
            fmt = impl_enum._map_DataTypeToStructFormatCode(self._metadata._ih._datatype)
            #print("_setBrickFilePosition: constant", type(content), content)
            tmp = struct.pack(fmt, content)
            tmp += bytes(8 - len(tmp))
            content = struct.unpack_from("<Q", tmp)[0]
            content |= (1<<63)
            self._metadata._blup._lookup[ix] = content
            self._metadata._blup._lookend[ix] = 0
        elif brickstatus == impl_enum.BrickStatus.Compressed:
            self._metadata._blup._lookend[ix] = content + bricksize
            content &= ~(0xFF << 56)
            content |= (0xC0 << 56)
            self._metadata._blup._lookup[ix] = content
        else:
            self._metadata._blup._lookup[ix] = content
            self._metadata._blup._lookend[ix] = content + bricksize

    def _getAlphaLookupIndex(self, i, j, lod):
        self._validatePosition(i, j, 0, lod)
        size = self._metadata._ih._lodsizes[lod]
        index = self._metadata._ih._alphaoffsets[lod] + i + (size[0] * j)
        if index < 0 or index >= self._metadata._ih._alphaoffsets[-1]:
            raise ZgyInternalError("Internal error in _getAlphaLookupIndex")
        return index

    def _getAlphaFilePosition(self, i, j, lod):
        # TODO-Compression, same handling as in _getBrickFilePosition()
        pos = self._metadata._alup._lookup[self._getAlphaLookupIndex(i, j, lod)]
        if pos == 0:
            return (impl_enum.BrickStatus.Missing, None, 0)
        elif pos == 1:
            return (impl_enum.BrickStatus.Constant, None, 0)
        elif pos & (1<<63):
            constant = pos & 0xff
            return (impl_enum.BrickStatus.Constant, None, constant)
        else:
            return (impl_enum.BrickStatus.Normal, pos, None)

    def _unscramble_init(self):
        """
        Create static tables to convert between a regular 3d brick (v2, v3)
        and a brick that has 8x8 subtiling in the horizontal direction (v1).
        """
        self._srctodst = np.zeros(64*64*64, dtype=np.int32)
        self._dsttosrc = np.zeros(64*64*64, dtype=np.int32)
        pos = 0
        for ii in range(0, 64, 8):
            for jj in range(0, 64, 8):
                for subi in range(ii, ii+8):
                    for subj in range(jj, jj+8):
                        for subk in range(64):
                            src_pos = pos
                            pos += 1
                            dst_pos = subi*64*64 + subj*64 + subk
                            self._srctodst[src_pos] = dst_pos
                            self._dsttosrc[dst_pos] = src_pos

    def _unscramble(self, brick):
        return brick.flat[self._srctodst].reshape(brick.shape)

    def readConstantValue(self, start, size, lod = 0, as_float = True, *, verbose = None):
        """
        Check to see if the specified region is known to have all
        samples set to the same value. A return value that is not None
        signifies that a regular read would return all samples set to
        that value. A return value of None means we don't know. This
        method is only intended as a hint to improve performance.
        """
        defaultstorage = self._defaultValue(as_float=False)
        needed = self._bricksNeeded(start, size, self._metadata._ih._bricksize)
        bricks = [((e[0],) + self._getBrickFilePosition(*e[1], lod)) for e in needed]
        result = None
        for startpos, brickstatus, fileoffset, constvalue, bricksize in bricks:
            if brickstatus == impl_enum.BrickStatus.Constant:
                if result is not None and result != constvalue and not (np.isnan(result) and np.isnan(constvalue)):
                    return None
                result = constvalue
            elif brickstatus == impl_enum.BrickStatus.Missing:
                if result is not None and result != defaultstorage:
                    return None
                result = defaultstorage
            else:
                return None
        if as_float:
            result = self._scaleDataToFloat(result)
        return result

    def _deliverOneBrick(self, result, start, startpos, raw, brickstatus, as_float, *, verbose = None):
        """
        This is the final step in readToExistingBuffer(). The data has
        been read from storage, so now it needs to be copied back to
        the user. This function may be invoked multiple times if data
        was needed from more than one brick.
        Arguments:
          result   -- user's buffer which was passed to readToExistingBuffer()
          start    -- user's supplied position as a 3-tuple of index values
          startpos -- position of the start of this brick
          raw      -- bulk data for this brick, scalar or bytes or ...
          brickstatus -- normal, compressed, constant, ...
        This low level function deals with bytes (or similar) and scalars
        an input. Not the np.ndarray and ScalarBuffer used elsewhere.
        """
        self._checkCompatibleBufferTypes(result.dtype, not as_float)
        file_dtype = impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype)
        # TODO-Low: Refactor: this is probably NOT the right place.
        if brickstatus == impl_enum.BrickStatus.Compressed:
            # Note that when the file contains compressed integral data
            # and the application asks for float data back, the decompressor
            # is responsible for the int to float conversion but NOT the
            # scaling from storage values to real values.
            onebrick = CompressFactoryImpl.decompress(raw, brickstatus, self._metadata._ih._bricksize, file_dtype, result.dtype)
            if onebrick is None:
                raise ZgyFormatError("Compression type not recognized")
        elif brickstatus == impl_enum.BrickStatus.Normal:
            if len(raw) != np.product(self._metadata._ih._bricksize) * np.dtype(file_dtype).itemsize:
                raise ZgyFormatError("Got wrong count when reading brick.")
            onebrick = np.frombuffer(raw, dtype=file_dtype)
            # Instead of describing the array as explicitly little-endian
            # I will immediately byteswap it to become native-endian.
            if np.dtype('<i4') != np.int32: # big-endian
                onebrick.byteswap(inplace=True)
            onebrick = onebrick.reshape(self._metadata._ih._bricksize, order="C")
            # Only needed for normal bricks. No compressed bricks exist in v1.
            if self._metadata._fh._version == 1:
                onebrick = self._unscramble(onebrick)
        else:
            _ = float(raw) # just to assert it is in fact a scalar.
            onebrick = ScalarBuffer(self._metadata._ih._bricksize, raw, file_dtype)
        if as_float and file_dtype != np.float32:
            onebrick = self._scaleDataToFloat(onebrick)
        #print(onebrick)

        # Note, we might pass in the survey range as well to force
        # all padding bytes to be set to the default value. Less
        # surprises for the caller. It may look a bit odd if the user
        # does a flood-fill of the entire survey to a given value and
        # later sees that the content is different in the padding area.
        # But, the caller should ignore the padding.
        #
        # On write the padding samples should also be forced to contain
        # the same value. If nothing else, to help compression. But for
        # efficiency reasons the value is not specified. Typically it
        # will be the default "absent" value but if the rest of the
        # brick has a const value and no allocated disk space then
        # they will inherit that constant.
        self._partialCopy(onebrick,
                          startpos, self._metadata._ih._bricksize,
                          result, start, result.shape,
                          (0, 0, 0), self._metadata._ih._size,
                          verbose=verbose)

    def readToExistingBuffer(self, result, start, lod, as_float, *, verbose = None):
        """
        Read bulk data starting at "start" in index space and store the
        result in the provided 3d numpy array. Start should be in the range
        (0,0,0) to Size-1. The count of samples to read is implied by the
        size of the provided result array that is passed in. The valid data
        types for the result array are float32 (in which case samples stored
        as int8 or int16 will be scaled) or the files's storage value type
        (in which case there is no scaling). It is valid to pass a count
        that includes the padding area between the survey and the end
        of the current brick, but not past that point.
        """
        self._argType(result, (np.ndarray,))
        self._checkCompatibleBufferTypes(result.dtype, not as_float)
        file_dtype = impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype)
        result_dtype = result.dtype
        count = result.shape

        # Need a default value to use when trying to read a brick that
        # was never written, or to fill in a brick that was only partly
        # written. To avoid non intuitive behavior the same value should
        # be used for both cases, and it should be a value that would
        # also be allowed to store as a regular sample. Use the value
        # that becomes 0.0 after conversion to float. Or as close as
        # possible to that value if the coding range is not zero centric.
        # Always use zero for floating point data. (Not NaN...)
        # Dead traces don't have any special handling apart from the
        # alpha flag. They still contain whatever data was written to them.
        defaultstorage = self._defaultValue(as_float=False)
        defaultvalue = self._defaultValue(as_float=True)

        # Make a separate pass to gather all the bricks we need to read.
        # FUTURE: we might fetch some of them in parallel and we might be
        # able to combine bricks to read larger blocks at a time. Both changes
        # can have a dramatic performance impact effect on cloud access.
        needed = self._bricksNeeded(start, count, self._metadata._ih._bricksize)
        bricks = [((e[0],) + self._getBrickFilePosition(*e[1], lod)) for e in needed]
        #print(bricks)

        #print("Default storage", defaultstorage, " and value", defaultvalue)
        result.fill(defaultvalue if _padding_fill_value is None else _padding_fill_value)
        # After all bricks have been processed, the padding past the
        # end if the survey might still not have been touched. Just in
        # case the request did in fact include such samples we will
        # initialize the entire result buffer to the default value.
        result.fill(_padding_fill_value if _padding_fill_value is not None else
                    defaultvalue if result.dtype == np.float32 else
                    defaultstorage)
        # TODO-High compression, use real_bricksize at least if compressed.
        bricksize_bytes = np.prod(self._metadata._ih._bricksize) * file_dtype().itemsize
        requests = []
        for startpos, brickstatus, fileoffset, constvalue, real_bricksize in bricks:
            # Functor that accepts a raw "data" brick and copies it
            # into the correct place in result. Note that "startpos" and
            # "brickstatus" need to be captured for the current loop
            # iteration. Other captured variables should remain the same.
            deliverance = lambda data, startpos=startpos, brickstatus=brickstatus: self._deliverOneBrick(result, start, startpos, data, brickstatus, as_float=as_float, verbose=verbose)
            if brickstatus == impl_enum.BrickStatus.Constant:
                if verbose:
                    verbose(" Reading brick at {0} as constant {1}".format(
                        startpos, constvalue))
                deliverance(constvalue)
            elif brickstatus == impl_enum.BrickStatus.Missing:
                if verbose:
                    verbose(" Reading brick at {0} not found, use {1}".format(
                        startpos, defaultstorage))
                deliverance(defaultstorage)
            elif brickstatus == impl_enum.BrickStatus.Normal:
                if verbose:
                    verbose("Reading brick at {0} from file offset {1}".format(
                        startpos, hex(fileoffset)))
                requests.append((fileoffset, bricksize_bytes, deliverance))
            elif brickstatus == impl_enum.BrickStatus.Compressed:
                if verbose:
                    verbose("Reading compressed brick at {0} from file offset {1} size {2}".format(
                        startpos, hex(fileoffset), hex(real_bricksize)))
                # TODO-Worry obscure corner case, might need to re-try if we didn't get enough data.
                requests.append((fileoffset, real_bricksize, deliverance))
            else:
                raise ZgyInternalError("Internal error, bad brick status")
        # Send all pending read requests to the backend in one operation.
        if requests:
            self._file.xx_readv(requests,
                                parallel_ok=False,
                                immutable_ok=False,
                                transient_ok=True,
                                usagehint=impl_file.UsageHint.Data)

    # --- WRITE SUPPORT --- #
    def _scaleDataToStorage(self, data):
        """
        Convert and/or scale the user-supplied data to the storage value type.
        Unlike _scaleToStorage() this is an instance method so it already
        knows the storage type. Also be more strict about the input type.
        The method accepts np.ndarray or ScalarBuffer only.
        """
        file_dtype = impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype)
        data_dtype = data.dtype

        if data_dtype != np.float32:
            raise ZgyInternalError("Input to _scaleDataToStorage must be np.float32, not " + str(data_dtype))
        if file_dtype != np.float32:
            data = self._scaleToStorage(data, self._metadata._ih._safe_codingrange, file_dtype)
        return data

    @classmethod
    def _setPaddingToEdge(cls, data, used, modulo, dim):
        """
        Pad unused parts of the data buffer by replicating the last samples,
        but only up to a multiple of 'modulo' samples. Handles just one
        dimension, so caller will typically invoke us three times.
        """
        cls._argType(data, (np.ndarray,))
        beg = used[dim]
        end = min(((used[dim]+modulo-1) // modulo) * modulo, data.shape[dim])
        if beg >= end: pass
        elif dim == 0: data[beg:end,:,:] = data[beg-1:beg,:,:]
        elif dim == 1: data[:,beg:end,:] = data[:,beg-1:beg,:]
        elif dim == 2: data[:,:,beg:end] = data[:,:,beg-1:beg]

    @classmethod
    def _setPaddingToConst(cls, data, used, missingvalue, dim):
        """
        Pad unused parts of the data buffer with a constant.
        Handles just one dimension, so caller should invoke us three times.
        """
        cls._argType(data, (np.ndarray,))
        beg = used[dim]
        end = data.shape[dim]
        if beg >= end: pass
        elif dim == 0: data[beg:end,:,:] = missingvalue
        elif dim == 1: data[:,beg:end,:] = missingvalue
        elif dim == 2: data[:,:,beg:end] = missingvalue

    @classmethod
    def _setPaddingSamples(cls, data, used, missingvalue, compressor):
        """
        Make sure the contants of the padding area, if any, is
        deterministic instead of whatever garbage the buffer holds.
        The method is allowed to update the buffer in place,
        but may also allocate and return a new one.
        """
        cls._argType(data, (np.ndarray,))

        if tuple(used) == tuple(data.shape):
            return data

        # TODO-Low let the compressor choose the optimal padding.
        # ZFP: replicate up to modulo-4, storage-zero outside.
        #if compressor: return compressor.pad(data, used)

        for dim in range(3):
            cls._setPaddingToConst(data, used, missingvalue, dim)

        # To reduce LOD edge artifacts, pad 'edge' up to even size.
        # To optimize ZFP compression, pad 'edge' up to modulo-4.
        for dim in range(3):
            cls._setPaddingToEdge(data, used, 4, dim)

        return data

    def _writeWithRetry(self, rawdata, brickstatus, fileoffset, brickpos, lod, *, verbose = None):
        dprint("@@ _writeWithRetry(pos={0}, count={1}, fileoffset={2:x})".format(brickpos, len(rawdata), fileoffset or self._file.xx_eof))
        if fileoffset is not None:
            try:
                self._file.xx_write(rawdata, fileoffset, usagehint=impl_file.UsageHint.Data)
            except ZgySegmentIsClosed:
                # The update mode doesn't need to be checked again here
                # unless we want to support UpdateMode.NoLeaks which
                # would require raising an exception.
                # TODO-Test: testing will require careful thought.
                fileoffset = None # Write a new brick, abandoning the old one.
        if fileoffset is None:
            # Write a new brick.
            fileoffset = self._file.xx_eof
            self._file.xx_write(rawdata, fileoffset, usagehint=impl_file.UsageHint.Data)
            self._setBrickFilePosition(brickpos[0], brickpos[1], brickpos[2],
                                       lod, brickstatus,
                                       fileoffset, len(rawdata))

    def _writeOneNormalBrick(self, data, fileoffset, brickpos, lod, compressor, *, verbose = None):
        """
        Handle padding of area outside the survey and optionally compression.
        Also convert from NumPy array to plain bytes.

        TODO-Multithreading

        This function is a candidate for running multi threaded due to the
        potentially expensive compression. While the next step, actually
        writing the file, needs to be serialized both because it may want
        to write at EOF and because even if the file offset is explicitly
        known the lower level write might not be thread safe. Also when
        serializing the order of bricks should be preserved. Otherwise
        performance on read might suffer.
        """
        self._argType(data, (np.ndarray,))
        dprint("@@ _writeOneNormalBrick(pos={0}, count={1})".format(brickpos, data.shape))
        # TODO-Medium: if scalar(data): return data, impl_enum.BrickStatus.Constant
        # Make it possible for this function to be called for all data
        # in _writeOneBrick(). Then move it even further up to
        # _writeAlignedRegion() which will hopefully enable some
        # parallelization on write.
        assert tuple(data.shape) == tuple(self._metadata._ih._bricksize)
        used = self._usedPartOfBrick(data.size, brickpos, lod)
        data = self._setPaddingSamples(data, used,
                                       missingvalue=self._defaultValue(as_float=False),
                                       compressor=compressor)
        # TODO-Compression, might also want to check for all-constant
        # by calling _isUsedPartOfBrickAllConstant() and if true, return
        # (data.flat[0], impl_enum.BrickStatus.Constant) and do not attempt
        # to compress. This function would then be called from
        # _writeAlignedRegion() instead of _writeWithRetry().
        # There is a slight amount of added work because if the data block
        # already exists on file (which we don't know yet) the test would
        # not be needed.
        # With that change both _writeOneNormalBrick() and _writeOneBrick()
        # will need to know the current brick status, and _writeOneNormalBrick()
        # would return .Constant for constant data.
        brickstatus = impl_enum.BrickStatus.Normal
        rawdata = None
        if compressor:
            cdata = compressor(data)
            if cdata:
                rawdata = cdata
                brickstatus = impl_enum.BrickStatus.Compressed
        if rawdata is not None:
            pass
        elif np.dtype('<i4') != np.int32: # big-endian, TODO-Worry not tested.
            rawdata = data.byteswap().tobytes()
        else:
            rawdata = data.tobytes()
        self._writeWithRetry(rawdata, brickstatus, fileoffset, brickpos, lod, verbose = None)

    def _writeOneConstantBrick(self, data, brickpos, lod, *, verbose = None):
        self._argType(data, (ScalarBuffer,))
        dprint("@@ _writeOneConstantBrick(pos={0}, value={1})".format(brickpos, data))
        self._setBrickFilePosition(brickpos[0], brickpos[1], brickpos[2], lod,
                                   impl_enum.BrickStatus.Constant, data.value, 0)

    def _usedPartOfBrick(self, size, brickpos, lod):
        """
        Compute the size of the used (inside survey) area of a data buffer
        with size "size". size will probably always be equal to bricksize
        but the function should still work if it isn't.
        """
        # TODO-Low: Refactor, should I enforce storing the members as numpy arrays?
        bricksize  = np.array(self._metadata._ih._bricksize, dtype=np.int64)
        surveysize = np.array(self._metadata._ih._size, dtype=np.int64)
        brickpos   = np.array(brickpos, dtype=np.int64)
        surveysize = (surveysize + (1<<lod) - 1) // (1<<lod)
        available =  surveysize - (brickpos * bricksize)
        return np.maximum((0, 0, 0), np.minimum(available, size))

    def _isUsedPartOfBrickAllConstant(self, data, brickpos, lod):
        """
        Return True if all useful samples in this brick have the same value.
        Padding samples outside the survey are not useful and should not
        be checked since they could easily have been set to a different value.
        """
        self._argType(data, (np.ndarray,))
        used = self._usedPartOfBrick(data.shape, brickpos, lod)
        first = data.flat[0]
        if np.isnan(first):
            return np.all(np.isnan(data[:used[0],:used[1],:used[2]]))
        else:
            return np.all(data[:used[0],:used[1],:used[2]] == first)

    def _mustLeakOldBrick(self, data, compressor, brickstatus):
        """
        Return True if this block needs to be leaked by pretending
        that its block offset has not been allocated yet.
        Raise an exception if the update us disallowed by _update_mode.

        Note, in "Always" mode I am only supposed to leak the block
        if the new one is larger. But that is too much work to
        figure out here. So, treat "Always" as if it were "Pedantic".

        Note, there is one other place where bricks might leak:
        in _writeWithRetry() when a ZgySegmentIsClosed is caught.
        The update mode doesn't need to be checked again at that point
        unless we want to support UpdateMode.NoLeaks.
        """
        self._argType(data, (np.ndarray, ScalarBuffer))
        msg = "Updating a {0} BrickStatus.{1} brick with {2} data is illegal in UpdateMode.{3}.".format(
            "cloud" if self._file.xx_iscloud else "local",
            brickstatus.name,
            "Compressed" if compressor else "Normal" if isinstance(data, np.ndarray) else "Constant",
            self._update_mode.name)
        if brickstatus != impl_enum.BrickStatus.Missing:
            if self._update_mode in (impl_enum.UpdateMode.Never,):
                raise ZgyUserError(msg)
        if brickstatus == impl_enum.BrickStatus.Normal:
            if self._update_mode in (impl_enum.UpdateMode.Never, impl_enum.UpdateMode.Constant):
                raise ZgyUserError(msg)
        if brickstatus == impl_enum.BrickStatus.Compressed or (brickstatus == impl_enum.BrickStatus.Normal and compressor is not None):
            if self._update_mode in (impl_enum.UpdateMode.Pedantic, impl_enum.UpdateMode.Always):
                return True
            else:
                raise ZgyUserError(msg)
        return False

    def _writeOneBrick(self, data, brickpos, lod, compressor, *, verbose = None):
        """
        Write a single brick to storage. Normally the data should either
        be a numpy scalar (all samples have the same value) or a numpy array
        with the same shape as the file's brick size.

        brickpos is given relative to this lod level. For lod 0 the valid
        range is the survey size in bricks. For lod 1 it is half that,
        rounded upwards.

        The data must already have been scaled to storage values and
        converted to the appropriate value type. It must either be a
        scalar (to fill the entire brick) or a numpy array.

        Caveat: int and float scalars won't work. Be explicit and pass an
        np.int8, np.int16, or np.float32.
        """
        self._argType(data, (np.ndarray, ScalarBuffer))
        self._checkCompatibleBufferTypes(data.dtype, is_storage = True)
        brickstatus, fileoffset, constvalue, bricksize = self._getBrickFilePosition(brickpos[0], brickpos[1], brickpos[2], lod)

        if self._mustLeakOldBrick(data, compressor, brickstatus):
            brickstatus, fileoffset, constvalue, bricksize = (
                impl_enum.BrickStatus.Missing, None, 0, 0)

        data_const = not isinstance(data, np.ndarray)
        file_const = brickstatus not in (impl_enum.BrickStatus.Normal, impl_enum.BrickStatus.Compressed)
        if file_const: # also true if never written yet.
            if data_const:
                # Caller asked to store a constant value.
                dprint("@@ Explicit store constant")
                self._writeOneConstantBrick(data, brickpos, lod, verbose=verbose)
            elif self._isUsedPartOfBrickAllConstant(data, brickpos, lod):
                # Caller did not explicitly ask for a constant value,
                # but all useable samples (i.e. all samples that are
                # inside the survey boundaries) have ths same value.
                tmp = ScalarBuffer(data.shape, data[0,0,0], data.dtype)
                self._writeOneConstantBrick(tmp, brickpos, lod, verbose=verbose)
            else:
                # Allocate a new disk block and write out this block.
                self._writeOneNormalBrick(data, None, brickpos, lod, compressor=compressor, verbose=verbose)
        else:
            if data_const:
                dprint("@@ Const data expanded before storing")
                # The brick has already been allocated. Cannot set it to
                # constant-value because the file storage for this brick
                # would then be leaked.
                data = np.full(self._metadata._ih._bricksize[:3], data.value, dtype=data.dtype)
            # No compression, for the same reason as above.
            self._writeOneNormalBrick(data, fileoffset, brickpos, lod, compressor=None, verbose=verbose)

    def _writeAlignedRegion(self, data, start, lod, compressor, *, verbose = None):
        """
        Write an arbitrary region covering one or more full bricks.
        The data must already be converted to storage.
        Start must be a multiple of brick size. start + count must
        either be a multiple of block size or set to the end of the
        survey.
        With current usage the data may or may not have been padded to
        the brick size depending on whether a read/modify/write was used.
        """
        self._argType(data, (np.ndarray, ScalarBuffer))
        # Meta information from the file only.
        bs = np.array(self._metadata._ih._bricksize, dtype=np.int64)
        file_dtype = impl_enum._map_DataTypeToNumpyType(self._metadata._ih._datatype)
        self._checkCompatibleBufferTypes(data.dtype, is_storage=True)
        survey_beg = np.array((0, 0, 0), dtype=np.int64)
        survey_end = np.array(self._metadata._ih._size, dtype=np.int64)
        survey_end = (survey_end + (1<<lod) - 1) // (1<<lod)
        defaultstorage = self._defaultValue(as_float=False)
        # Massaging of arguments to this function
        dprint("@@ _writeRegion(start {0}, count {1}, type {2} -> {3})".format(
            start, data.shape, data.dtype, np.dtype(file_dtype)))
        start = np.array(start, dtype=np.int64)
        count = np.array(data.shape, dtype=np.int64)
        beg_brick = (start // bs) * bs
        end_brick = ((start + count + bs - 1) // bs) * bs
        if isinstance(data, np.ndarray):
            brick = np.zeros(bs, dtype=file_dtype)
        else:
            brick = ScalarBuffer(bs, data.value, data.dtype)
        for ii in range(beg_brick[0], end_brick[0], bs[0]):
            for jj in range(beg_brick[1], end_brick[1], bs[1]):
                for kk in range(beg_brick[2], end_brick[2], bs[2]):
                    this_beg = np.array([ii, jj, kk], dtype=np.int64)
                    brickpos = this_beg // bs
                    if isinstance(brick, np.ndarray):
                        brick.fill(defaultstorage)
                        self._partialCopy(data, start, data.shape,
                                          brick, this_beg, bs,
                                          survey_beg, survey_end - survey_beg,
                                          verbose=verbose)
                    # TODO-Medium Note Compression:
                    # This might be a good extension point, as this is the
                    # last place where we still have a list of bricks to
                    # be written. So it would be easier to parallelize here.
                    # But we don't yet know which bricks will be written as
                    # all-constant so there would need to be some refactoring.
                    # _writeOneNormalBrick() probably needs to be called from here.
                    # See comments in that function; it will need some
                    # changes. And _writeOneBrick() needs to be told whether
                    # the data has been compressed or not. Instead of
                    # being told which compressor (if any) to use.

                    # If there are any errors during _writeOneBrick() this
                    # probably means the entire file is a lost cause.
                    # This is true also for ZgyUserError raised in the file
                    # layer, because at that layer the "user" is really
                    # OpenZGY and not some client code. The only acceptable
                    # error is ZgySegmentIsClosed, and that will be caught
                    # and handled at lower levels.

                    # TODO-Low the logic here can probably be refined.
                    # If there is no buffering (on-prem files) then a
                    # write can probably just be re-tried if we make
                    # sure the write is done before updating metadata.
                    # But do we really want that complexity? The risk
                    # of transient errors is way higher with cloud access.
                    # And due to buffering in the file layer those would
                    # also be a lot harder to recover from.

                    with ErrorsWillCorruptFile(self):
                        self._writeOneBrick(brick, brickpos,
                                            lod=lod,
                                            compressor=compressor,
                                            verbose=verbose)

    def _writeRegion(self, data, start, lod, compressor, is_storage, *, verbose = None):
        """
        Write an arbitrary region. If start and end are not aligned to
        the brick size this will do a read/modify/write. Note that when
        writing to the cloud it is highly recommended to write aligned
        regions only. Otherwise some disk space might be wasted.

        Performance note: The read/modify/write could also have been done
        one brick at a time. Doing it here means that for large requests
        a number of bricks which were already full will also be read.
        On the other hand a single read might help parallelism. Either
        way the recommendation is for the user to write brick aligned data.
        So this might not be a big deal.
        """
        self._argType(data, (np.ndarray, ScalarBuffer))
        self._checkCompatibleBufferTypes(data.dtype, is_storage)

        if not is_storage:
            data = self._scaleDataToStorage(data)

        beg = np.array(start, dtype=np.int64)
        end = np.array(data.shape, dtype=np.int64) + beg
        bs = np.array(self._metadata._ih._bricksize, dtype=np.int64)
        survey_beg = np.array((0, 0, 0), dtype=np.int64)
        survey_end = (np.array(self._metadata._ih._size, dtype=np.int64) + ((1<<lod) - 1)) // (1<<lod)
        need_rmw = any([(beg[i]%bs[i]) != 0 or ((end[i]%bs[i]) != 0 and end[i] < survey_end[i]) for i in range(3)])
        if need_rmw:
            # Since we are doing a read/modify/write, also read any padding
            # outside the survey. _writeAlignedRegion will see full
            # bricks also at the end of the survey. If no r/m/w needed
            # the latter is not guaranteed.
            new_start = (beg // bs) * bs
            new_count = (((end + bs - 1) // bs) * bs) - new_start
            new_data = np.zeros(new_count, dtype=data.dtype)
            dprint("@@ _writeRegion r/m/w: user {0} {1} padded {2} {3}".format(
                start, data.shape, new_start, new_count))
            # The reader takes care of any defaultvalue.
            # as_float=False because we converted, and data.dtype==file_dtype
            self.readToExistingBuffer(new_data, new_start, lod=lod,
                                      as_float=False,
                                      verbose=verbose)
            self._partialCopy(data, start, data.shape,
                              new_data, new_start, new_data.shape,
                              survey_beg, survey_end - survey_beg,
                              verbose=verbose)
            data = new_data
            start = new_start
        else:
            dprint("@@ _writeRegion direct: user {0} {1}".format(start, data.shape))

        self._writeAlignedRegion(data, start, lod=lod,
                                 compressor=compressor,
                                 verbose=verbose)

        data = data.value if isinstance(data, ScalarBuffer) else data
        if np.issubdtype(data.dtype, np.integer):
            tmp_min = np.amin(data)
            tmp_max = np.amax(data)
            if self._sample_min > self._sample_max:
                # Changes _sample_min / _sample_max to the correct value type.
                self._sample_min = tmp_min
                self._sample_max = tmp_max
            else:
                self._sample_min = min(self._sample_min, tmp_min)
                self._sample_max = max(self._sample_max, tmp_max)
        else:
            valid = np.isfinite(data).astype(np.bool)
            self._sample_min = np.amin(data, where=valid, initial=self._sample_min)
            self._sample_max = np.amax(data, where=valid, initial=self._sample_max)

# Copyright 2017-2021, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



"""Open ZGY Wrapper Module for Xarray"""

# look for local installation
import sys
import importlib
import pathlib

_missing_local = True
_missing_submodule = True

try:
    import openzgy

    _missing_local = False
except ImportError:
    pass

if _missing_local:
    _here = pathlib.Path(__file__).absolute().parent
    _finder = importlib.machinery.FileFinder(str(_here / "open-zgy/python"))
    _spec = _finder.find_spec("openzgy")
    if _spec is not None:
        _module = importlib.util.module_from_spec(_spec)
        sys.modules["openzgy"] = _module
        _spec.loader.exec_module(_module)
        _missing_submodule = False

if _missing_local and _missing_submodule:
    raise ImportError(
        "The OpenZgy Module requires that Open ZGY for Python"
        " be installed. Checkout the module ReadMe for more information."
    )

from openzgy.api import ZgyReader, ZgyWriter, SampleDataType, UnitDimension

import numpy as np
from .._seismic_dataset import create3d_dataset
from .._keyfield import CoordKeyField, VariableKeyField, AttrKeyField
from ..segy._segy_writer import _check_dimension


def zgy_loader(filename):
    """Load a ZGY File.

    Args:
        filename (pathlib.Path/str): The path or file location of the file to load.

    Returns:
        xarray.Dataset: Loaded data.
    """
    with ZgyReader(str(filename)) as reader:
        header = reader.meta

        data_shape = header["size"]
        zstart = header["zstart"]
        zinc = header["zinc"]
        # zunit_fact = header["zunitfactor"] may need this later
        ilstep, xlstep = header["annotinc"]
        ilstart, xlstart = header["annotstart"]

        corners = header["corners"]

        ds = create3d_dataset(
            data_shape,
            first_sample=zstart,
            sample_rate=zinc,
            iline_step=ilstep,
            xline_step=xlstep,
            first_iline=ilstart,
            first_xline=xlstart,
        )

        # load attributes
        ds.attrs[AttrKeyField.corner_points_xy] = [c for c in corners]
        ds.attrs[AttrKeyField.source_file] = str(filename)
        ds.attrs[AttrKeyField.text] = "SEGY-SAK Created from ZGY Loader"
        ds.attrs[AttrKeyField.ns] = data_shape[2]
        ds.attrs[AttrKeyField.coord_scalar] = header["hunitfactor"]

        # load data
        data = np.zeros(data_shape, dtype=np.float32)
        reader.read((0, 0, 0), data)
        ds[VariableKeyField.data] = (
            (CoordKeyField.iline, CoordKeyField.xline, "twt"),
            data,
        )

        # create XY world coordinates.
        ds[CoordKeyField.cdp_x] = (
            (CoordKeyField.iline, CoordKeyField.xline),
            np.zeros(data_shape[0:2]),
        )
        ds[CoordKeyField.cdp_y] = ds[CoordKeyField.cdp_x].copy()

        pos = np.dstack(
            [
                ds.iline.broadcast_like(ds[CoordKeyField.cdp_x]).values,
                ds.xline.broadcast_like(ds[CoordKeyField.cdp_x]).values,
            ]
        )

        pos = np.apply_along_axis(reader.annotToWorld, -1, pos)

        ds[CoordKeyField.cdp_x] = (
            (CoordKeyField.iline, CoordKeyField.xline),
            pos[:, :, 0],
        )
        ds[CoordKeyField.cdp_y] = (
            (CoordKeyField.iline, CoordKeyField.xline),
            pos[:, :, 1],
        )

        ds.seis.calc_corner_points()

        return ds


def zgy_writer(seisnc_dataset, filename, dimension=None):
    """Write a seisnc dataset to ZGY file format.

    Args:
        seisnc_dataset (xr.Dataset): This should be a seisnc dataset.
        filename (pathlib.Path/str): The filename to write to.
        dimension (str, optional): The dimension to write out, if
            None uses available dimenions twt first.
    """
    assert seisnc_dataset.seis.is_3d()

    seisnc_dataset.seis.calc_corner_points()
    corners = tuple(seisnc_dataset.attrs["corner_points_xy"][i] for i in [0, 3, 1, 2])

    dimension = _check_dimension(seisnc_dataset, dimension)
    if dimension == "twt":
        zunitdim = UnitDimension(2001)
    elif dimension == "depth":
        zunitdim = UnitDimension(2002)
    else:
        zunitdim = UnitDimension(2000)
    try:
        coord_scalar_mult = seisnc_dataset.attrs["coord_scalar_mult"]
    except KeyError:
        coord_scalar_mult = 1.0

    # dimensions
    ni, nj, nk = (
        seisnc_dataset.dims[CoordKeyField.iline],
        seisnc_dataset.dims[CoordKeyField.xline],
        seisnc_dataset.dims[dimension],
    )

    # vertical
    z0 = int(seisnc_dataset[dimension].values[0])
    dz = int(seisnc_dataset.sample_rate)

    # annotation
    il0 = seisnc_dataset[CoordKeyField.iline].values[0]
    xl0 = seisnc_dataset[CoordKeyField.xline].values[0]
    dil = int(np.median(np.diff(seisnc_dataset[CoordKeyField.iline].values)))
    dxl = int(np.median(np.diff(seisnc_dataset[CoordKeyField.xline].values)))

    dtype_float = SampleDataType["float"]

    with ZgyWriter(
        str(filename),
        size=(ni, nj, nk),
        datatype=dtype_float,
        zunitdim=zunitdim,
        zstart=z0,
        zinc=dz,
        annotstart=(il0, xl0),
        annotinc=(dil, dxl),
        corners=corners,
    ) as writer:

        order = (CoordKeyField.iline, CoordKeyField.xline, dimension)

        writer.write(
            (0, 0, 0),
            seisnc_dataset[VariableKeyField.data]
            .transpose(*order)
            .values.astype(np.float32),
        )


# TODO: Create a converter which does just a few ilines at a time. For large files.



"""Command line script for interacting with SEG-Y data.
"""

import os
import logging
import pathlib
import click
from tqdm import tqdm

try:
    from .version import version as VERSION
except ImportError:
    VERSION = None

if VERSION is None:
    try:
        from setuptools_scm import get_version

        VERSION = get_version(root="..", relative_to=__file__)
    except LookupError:
        VERSION = "¯\_(ツ)_/¯"

from segysak.segy import (
    segy_converter,
    segy_writer,
    segy_header_scan,
    segy_header_scrape,
    get_segy_texthead,
)
from segysak.tools import fix_bad_chars

# configuration setup
NAME = "segysak"
LOGGER = logging.getLogger(NAME)


def check_file(input_files):
    """
    Function expands a list of files if * wildcards are present. It also verifies the presence of each file.
    Depreciated since version 0.1.2: The CLI only accepts a single input file and Click performs the verification.
    """

    if input_files is None:
        LOGGER.error("Require input file/s")
        raise SystemExit

    checked_files = list()

    expanded_input = input_files.copy()

    for ifile in input_files:
        if "*" in ifile:
            ifile_path = pathlib.Path(ifile)
            parent = ifile_path.absolute().parent
            expanded_input += list(parent.glob(ifile_path.name))
            expanded_input.remove(ifile)

    for ifile in expanded_input:
        ifile = pathlib.Path(ifile)
        if ifile.exists():
            checked_files.append(ifile)
        else:
            LOGGER.error("Cannot find input {segyfile}")
            raise SystemExit

    return checked_files


def _action_ebcidc_out(arg, input_file):
    try:
        ebcidc = get_segy_texthead(input_file)
    except IOError:
        LOGGER.error("Input SEG-Y file was not found - check name and path")
        raise SystemExit

    if arg is None:
        print(ebcidc)
    else:
        ebcidc = fix_bad_chars(ebcidc)
        with open(arg, "w") as f:
            f.write(ebcidc)


def _action_ebcidc_in(input_ebcidc_file, input_file):
    pass


def guess_file_type(file):
    """
    Guess the file type. Currently guessing is only based on file extension.
    """
    _, file_extension = os.path.splitext(file)
    if any([ext in file_extension.upper() for ext in ["SEGY", "SGY"]]):
        return "SEGY"
    elif "SEISNC" in file_extension.upper():
        return "NETCDF"
    else:
        return None


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option(
    "--version",
    "-v",
    is_flag=True,
    help="Print application version name",
    default=False,
)
def cli(version):
    """
    The SEG-Y Swiss Army Knife (segysak) is a tool for managing segy data.
    It can read and dump ebcidc headers, scan trace headers, convert SEG-Y to SEISNC and vice versa
    """
    LOGGER.info(f"segysak v{VERSION}")
    if version:
        click.echo(f"{NAME} {VERSION}")
        raise SystemExit


@cli.command(help="Print SEG-Y EBCIDC header")
@click.argument("filename", type=click.Path(exists=True))
def ebcidc(filename):
    input_file = pathlib.Path(filename)
    click.echo(get_segy_texthead(input_file))


@cli.command(help="Scan trace headers and print value ranges")
@click.option(
    "--max-traces", "-m", type=click.INT, default=1000, help="Number of traces to scan"
)
@click.argument("filename", type=click.Path(exists=True))
def scan(max_traces, filename):
    input_file = pathlib.Path(filename)
    hscan = segy_header_scan(input_file, max_traces_scan=max_traces)
    click.echo(f"Traces scanned: {hscan.nscan}")
    import pandas as pd

    pd.set_option("display.max_rows", hscan.shape[0])
    click.echo(hscan[["byte_loc", "min", "max", "mean"]])


@cli.command()
@click.option(
    "--ebcidc", "-e", is_flag=True, default=False, help="Output the text header"
)
@click.option(
    "--trace-headers",
    "-h",
    is_flag=True,
    default=False,
    help="Output the trace headers to csv",
)
@click.argument("filename", nargs=-1, type=click.Path(exists=True))
def scrape(filename, ebcidc=False, trace_headers=False):
    """Scrape the file meta information and output it to text file.

    If no options are specified both will be output. The output file will be
    <filename>.txt for the EBCIDC and <filename>.csv for
    trace headers.

    The trace headers can be read back into Python using
    pandas.read_csv(<filename>.csv, index_col=0)
    """
    for file in tqdm(filename, desc="File"):
        file = pathlib.Path(file)
        ebcidc_name = file.with_suffix(".txt")
        header_name = file.with_suffix(".csv")

        if ebcidc == False and trace_headers == False:
            ebcidc = True
            trace_headers = True

        if ebcidc:
            txt = get_segy_texthead(file)
            with open(ebcidc_name, "w") as txtfile:
                txtfile.writelines(txt)

        if trace_headers:
            head_df = segy_header_scrape(file)
            head_df.to_csv(header_name)


@cli.command(
    help="Convert file between SEG-Y and NETCDF (direction is guessed or can be made explicit with the --output-type option)"
)
@click.argument(
    "input-files",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--output-file", "-o", type=click.STRING, help="Output file name", default=None
)
@click.option(
    "--iline", "-il", type=click.INT, default=189, help="Inline byte location"
)
@click.option(
    "--xline", "-xl", type=click.INT, default=193, help="Crossline byte location"
)
@click.option("--cdpx", "-x", type=click.INT, default=181, help="CDP X byte location")
@click.option("--cdpy", "-y", type=click.INT, default=185, help="CDP Y byte location")
@click.option(
    "--crop",
    type=click.INT,
    nargs=4,
    default=None,
    help="Crop the input volume providing 4 parameters: minil maxil minxl maxxl",
)
@click.option(
    "--output-type",
    type=click.Choice(["SEG-Y", "NETCDF"], case_sensitive=False),
    default=None,
    help="Explicitly state the desired output file type by chosing one of the options",
)
@click.option(
    "--dimension",
    "-d",
    type=click.STRING,
    default=None,
    help="Data dimension (domain) to write out, will default to TWT or DEPTH. Only used for writing to SEG-Y.",
)
def convert(
    output_file, input_files, iline, xline, cdpx, cdpy, crop, output_type, dimension
):

    if len(input_files) > 1 and output_file is not None:
        raise ValueError(
            "The output file option should not be used with multiple input files."
        )

    for input_file in input_files:
        input_file = pathlib.Path(input_file)
        if output_type is None and output_file is not None:
            output_type = guess_file_type(output_file)
        elif output_type is None and output_file is None:
            """Because currently only one conversion exists we can guess the output from the input"""
            input_type = guess_file_type(input_file)
            if input_type:
                output_type = "SEGY" if input_type == "NETCDF" else "NETCDF"

        if output_type is None:
            click.echo(
                "Output type not recognised! Please provide the desired output file type explicitly using the --output-type option"
            )
            raise SystemExit

        click.echo(f"Converting file {input_file.name} to {output_type}")

        if len(crop) == 0:
            crop_loc = None
        else:
            crop_loc = crop

        if output_type == "NETCDF":
            if output_file is None:
                output_file_loc = input_file.stem + ".SEISNC"
            else:
                output_file_loc = output_file

            segy_converter(
                input_file,
                ncfile=output_file_loc,
                iline=iline,
                xline=xline,
                ix_crop=crop_loc,
                cdpx=cdpx,
                cdpy=cdpy,
            )
            click.echo(f"Converted file saved as {output_file_loc}")
            LOGGER.info(f"NetCDF output written to {output_file_loc}")
        elif output_type == "SEG-Y":
            if output_file is None:
                output_file_loc = input_file.stem + ".segy"
            else:
                output_file_loc = output_file

            cdp_x = cdpx
            cdp_y = cdpy
            vars = locals()

            trace_header_map = {
                key: vars[key]
                for key in ["iline", "xline", "cdp_x", "cdp_y"]
                if vars[key] is not None
            }
            segy_writer(
                input_file,
                output_file_loc,
                trace_header_map=trace_header_map,
                dimension=dimension,
            )
            click.echo(f"Converted file saved as {output_file_loc}")
            LOGGER.info(f"SEG-Y output written to {output_file_loc}")
        else:
            click.echo(
                f"Conversion to output-type {output_type} is not implemented yet"
            )
            raise SystemExit


if __name__ == "__main__":
    cli()



import os
import datetime
from scipy.interpolate import interp1d
import numpy as np


def _check_crop(crop, limits):
    """Make sure a pair of limits is cropped correctly.

    Args:
        crop (list): Crop range [min, max]
        limits (list): Limits to crop [min, max]
    """
    if crop[0] > crop[1]:
        raise ValueError(f"Crop min: '{crop[0]}' is greater than crop max: '{crop[1]}'")
    if limits[0] > limits[1]:
        raise ValueError(
            f"Limits min: '{limits[0]}' is greater than limits max: '{limits[1]}'"
        )
    if limits[0] > crop[1] or limits[1] < crop[0]:
        raise ValueError(f"Crop range is outside data limits")


def _crop(crop, limits):
    """Perform cropping operation.

    Args:
        crop (list): Cropping limits [min, max]
        limits (list): Limits to crop [min, max]

    Returns:
        list: Cropped limits [min, max]
    """
    checked = limits.copy()
    if limits[0] < crop[0] < limits[1]:
        checked[0] = crop[0]
    if limits[0] < crop[1] < limits[1]:
        checked[1] = crop[1]
    return checked


def check_crop(crop, limits):
    """Make sure mins and maxs or horizontal cropping box are good.

    Args:
        crop (list): Input crop [min_il, max_il, min_xl, max_xl]
        limits (list): Input data limits [min_il, max_il, min_xl, max_xl]

    Returns:
        list: cropped limits [min_il, max_il, min_xl, max_xl]
    """
    checked = limits.copy()

    # check crop
    _check_crop(crop[:2], limits[:2])
    _check_crop(crop[2:], limits[2:])

    # crop
    checked[:2] = _crop(crop[:2], limits[:2])
    checked[2:] = _crop(crop[2:], limits[2:])

    return checked


def check_zcrop(crop, limits):
    """Make sure mins and maxs of vertical cropping box are good.

    Args:
        crop (list): Input crop [min, max]
        limits (list): Input data limits to crop [min, max]

    Output:
        list: cropped limits
    """
    _check_crop(crop, limits)
    return _crop(crop, limits)


def fix_bad_chars(text, sub="#"):
    """Remove unencodeable characters. And replace them with sub.

    Args:
        text (str): Text string to fix.
        sub (str): Sting to replace bad chars.

    Returns:
        (str): Cleaned text string.
    """
    good_text = ""
    for i, char in enumerate(text):
        if len(char.encode()) == 1:
            good_text = good_text + char
        else:
            good_text = good_text + "#"
    return good_text


def _get_datetime():
    """Return current date and time as formatted strings

    Returns:
        str, str: Date "YYYY-MM-DD" Time "HH:MM:SS"
    """
    now = datetime.datetime.now()
    return now.strftime("%Y-%m-%d"), now.strftime("%H:%M:%S")


def _get_userid():
    """Try to get the user id for reporting.

    Returns:
        str: The current user login ID.
    """
    try:
        return os.getlogin()
    except OSError:  # couldn't get it this way
        pass
    return "segysak_user"


def get_uniform_spacing(cdp_x, cdp_y, extra=None, bin_spacing_hint=10, method="linear"):
    """Interpolate the cdp_x, cdp_y arrays uniformly while staying close to the
    requested bin spacing

    Args:
        cdp_x (array-like)
        cdp_y (array-like)
        bin_spacing_hint (number): a bin spacing to stay close to, in cdp world units. Default: 10

    Returns:
        ndarray, ndarray, number: interpolated cdp_x, cdp_y and the actual bin spacing
    """
    xy = np.array([cdp_x, cdp_y])

    path_pos = np.insert(np.linalg.norm(np.diff(xy, axis=1), axis=0), 0, 0.0)

    path_length = path_pos.sum()
    segments = np.cumsum(path_pos)

    num_pts = int(path_length / bin_spacing_hint)
    uniform_sampled_path = np.linspace(0, segments[-1], num_pts)

    cdp_x_i = interp1d(segments, cdp_x, kind="linear")(uniform_sampled_path)
    cdp_y_i = interp1d(segments, cdp_y, kind="linear")(uniform_sampled_path)

    extras_i = {
        key: interp1d(segments, ex, kind="linear")(uniform_sampled_path)
        for key, ex in extra.items()
    }

    return cdp_x_i, cdp_y_i, uniform_sampled_path, extras_i


def halfsample(arr):
    """Sample an array at exactly half spacing.

    Uses the diff betweem consecutive samples to work out the
    inbetween samples.

    Args:
        arr (array-like): The sequence to resample.

    Returns:
        ndarray: The halfsampled array with (N*2 - 1) samples.
    """
    arr = np.atleast_1d(arr).astype(float)
    arr_halved = np.zeros(arr.size * 2 - 1)
    arr_halved[0::2] = arr
    arr_halved[1::2] = np.diff(arr) / 2.0 + arr[:-1]
    return arr_halved


def plane(xy, a, b, c):
    x, y = xy
    return a * x + b * y + c



#!/usr/bin/env python3

import numpy as np
import os, sys
import argparse
from ..api import ZgyReader, SampleDataType
from ..test.utils import SDCredentials

_brief_info = """
File name                      = '{name}'
File size (bytes)              = {r._fd.xx_eof:,d}
File format and version        = {r.datatype.name} ZGY version {r._accessor._metadata._fh._version}
Current data Version           = {r.verid}
Brick size I,J,K               = {r.bricksize}
Number of bricks I,J,K         = {r.brickcount[0]}
Number of LODs                 = {r.nlods}
Coding range min/max           = {r.datarange[0]:.6g} {r.datarange[1]:.6g} (raw: {r.raw_datarange[0]:.6g} {r.raw_datarange[1]:.6g}) {nsamples:,d}
Statistical min/max/count      = {r.statistics.min:.6g} {r.statistics.max:.6g} {r.statistics.cnt:,d}
Histogram range min/max/count  = {r.histogram.min:.6g} {r.histogram.max:.6g} {r.histogram.cnt:,d}
Inline start/increment/count   = {r.annotstart[0]} {r.annotinc[0]} {r.size[0]}
Xline  start/increment/count   = {r.annotstart[1]} {r.annotinc[1]} {r.size[1]}
Sample start/increment/count   = {r.zstart} {r.zinc} {r.size[2]}
Horizontal projection system   = {r._accessor._metadata._ih._hprjsys}
Horizontal dim/factor/name     = {r.hunitdim.name} {r.hunitfactor} '{r.hunitname}'
Vertical dim/factor/name       = {r.zunitdim.name} {r.zunitfactor} '{r.zunitname}'
Ordered Corner Points Legend   = [  <i>,   <j>] {{ <inline>,   <xline>}} (  <easting>,  <northing>)
Ordered Corner Point 1         = [{r.indexcorners[0][0]:5d}, {r.indexcorners[0][1]:5d}] {{{r.annotcorners[0][0]:9g}, {r.annotcorners[0][1]:9g}}} ({r.corners[0][0]:11.2f}, {r.corners[0][1]:11.2f})
Ordered Corner Point 2         = [{r.indexcorners[1][0]:5d}, {r.indexcorners[1][1]:5d}] {{{r.annotcorners[1][0]:9g}, {r.annotcorners[1][1]:9g}}} ({r.corners[1][0]:11.2f}, {r.corners[1][1]:11.2f})
Ordered Corner Point 3         = [{r.indexcorners[2][0]:5d}, {r.indexcorners[2][1]:5d}] {{{r.annotcorners[2][0]:9g}, {r.annotcorners[2][1]:9g}}} ({r.corners[2][0]:11.2f}, {r.corners[2][1]:11.2f})
Ordered Corner Point 4         = [{r.indexcorners[3][0]:5d}, {r.indexcorners[3][1]:5d}] {{{r.annotcorners[3][0]:9g}, {r.annotcorners[3][1]:9g}}} ({r.corners[3][0]:11.2f}, {r.corners[3][1]:11.2f})
"""

_hist_info = "Histogram bin {0:3d}              = {1:11d}"

def all_brick(reader):
    for lod in range(reader.nlods):
        for ii in range(reader.brickcount[lod][0]):
            for jj in range(reader.brickcount[lod][1]):
                for kk in range(reader.brickcount[lod][2]):
                    info = reader._accessor._getBrickFilePosition(ii,jj,kk,lod)
                    yield ((lod, ii, jj, kk),) + info

def all_alpha(reader):
    for lod in range(reader.nlods):
        for ii in range(reader.brickcount[lod][0]):
            for jj in range(reader.brickcount[lod][1]):
                info = reader._accessor._getAlphaFilePosition(ii,jj,lod)
                yield ((lod, ii, jj),) + info + (0,)

def summary_brick_offsets(reader):
    alpha = dict()
    brick = dict()
    for info in all_alpha(reader):
        alpha[info[1].name] = alpha.get(info[1].name, 0) + 1
    for info in all_brick(reader):
        brick[info[1].name] = brick.get(info[1].name, 0) + 1
    print("{0:30s} = {1}".format("Alpha status", str(alpha)))
    print("{0:30s} = {1}".format("Brick status", str(brick)))

def summary_normal_size(reader, *, header = True):
    """
    Useful for performance measurements.
    1) The number of allocated LOD0 bricks, used to compute bandwidth after
       timing the read of the entire file. Specified as bricks and MB.
    2) Size of one brick_column, allocated or not, given as above.
    3) Brick size in bytes.
    Note that size in MB is computed by integer division.
    If you need them exact, do the (float) division yourself.
    """
    x = [e for e in all_brick(reader) if e[0][0] == 0 and e[1].name == "Normal"]
    normal = len(x)
    bytespersample = {SampleDataType.int8:  1,
                      SampleDataType.int16: 2,
                      SampleDataType.float: 4}[reader.datatype]
    bytesperbrick = np.product(reader.bricksize) * bytespersample
    colsize = (reader.size[2] + reader.bricksize[2] - 1) // reader.bricksize[2]
    bytespercolumn = bytesperbrick * colsize
    fmt = "{0:30s} = LOD0: {1} {2} MB column {3} {4} MB brick {5}"
    if not header: fmt = "{1} {2} {3} {4} {5}"
    print(fmt.format(
        "Normal LOD0 bricks & col size", normal,
        (normal * bytesperbrick) // (1024*1024),
        colsize,
        (colsize * bytesperbrick) // (1024*1024),
        bytesperbrick))

def dump_brick_offsets(reader, sort):
    print("BRICK offsets:")
    table = all_brick(reader)
    if sort:
        table = sorted(table, key=lambda x: x[2] or 0)
    for (pos, brickstatus, fileoffset, constvalue, bricksize) in table:
        addr = "[{0}][{1}][{2}][{3}]".format(pos[0], pos[1], pos[2], pos[3])
        if fileoffset is not None:
            print("{addr:20s} = {fileoffset:16x} {brickstatus.name} Size {bricksize:8x}".format(
                addr=addr,
                brickstatus=brickstatus, fileoffset=fileoffset,
                constvalue=constvalue, bricksize=bricksize))
        else:
            print("{addr:20s} = {fileoffset:16s} {brickstatus.name} {constvalue}".format(
                addr=addr,
                brickstatus=brickstatus, fileoffset="",
                constvalue=constvalue, bricksize=""))

def dump_alpha_offsets(reader, sort):
    print("ALPHA offsets:")
    table = all_alpha(reader)
    if sort:
        table = sorted(table, key=lambda x: x[2] or 0)
    for (pos, brickstatus, fileoffset, constvalue, bricksize) in table:
        addr = "[{0}][{1}][{2}]".format(pos[0], pos[1], pos[2])
        if fileoffset is not None:
            print("{addr:20s} = {fileoffset:16x} {brickstatus.name}".format(
                addr=addr,
                brickstatus=brickstatus, fileoffset=fileoffset,
                constvalue=constvalue))
        else:
            print("{addr:20s} = {fileoffset:16s} {brickstatus.name} {constvalue}".format(
                addr=addr,
                brickstatus=brickstatus, fileoffset="",
                constvalue=constvalue))

def dump_combined_offsets(reader, sort):
    print("BRICK and ALPHA offsets sorted by address:")
    table = list(all_brick(reader)) + list(all_alpha(reader))
    if sort:
        table.sort(key=lambda x: x[2] or 0)
    for (pos, brickstatus, fileoffset, constvalue, bricksize) in table:
        if len(pos) == 4:
            addr = "brick [{0}][{1}][{2}][{3}]".format(pos[0], pos[1], pos[2], pos[3])
        elif len(pos) == 3:
            addr = "alpha [{0}][{1}][{2}]".format(pos[0], pos[1], pos[2])
        if fileoffset is not None:
            print("{addr:26s} = {fileoffset:16x} {brickstatus.name} Size {bricksize:8x}".format(
                addr=addr,
                brickstatus=brickstatus, fileoffset=fileoffset,
                constvalue=constvalue, bricksize=bricksize))
        else:
            print("{addr:26s} = {fileoffset:16s} {brickstatus.name} {constvalue}".format(
                addr=addr,
                brickstatus=brickstatus, fileoffset="",
                constvalue=constvalue, bricksize=""))

def run(filename, options):
    with ZgyReader(filename, iocontext = SDCredentials()) as reader:
        if options.only_lod0_info:
            summary_normal_size(reader, header=False)
            return
        args = dict(name=filename,
                    nsamples=np.product(reader.size),
                    r=reader)
        #print(_brief_info.format(**args))
        for line in _brief_info.split('\n'):
            if line:
                try:
                    print(line.format(**args))
                except Exception as ex:
                    print(line.split('=')[0] + "= N/A " + str(ex))
        summary_brick_offsets(reader)
        summary_normal_size(reader)
        if options.histogram:
            hh = reader.histogram.bin
            for ii in range(len(hh)):
                print(_hist_info.format(ii, hh[ii]))
        if options.sorted_offsets:
            dump_combined_offsets(reader, True)
        elif options.offsets:
            dump_brick_offsets(reader, False)
            dump_alpha_offsets(reader, False)

# Features from the old C++ zgydump
#    -b --brief BriefInfo()
#    (default)  FullInfo()
#       In new code, --histogram must be requested explicitly.
#       There is no --brief since the two outputs then end up
#       rather similar.
#    -p --performance PerfInfo()
#       normal/empty/const counts now printed unconditionally.
#       Can also consider merging test_zgydump and test_isoptimal.
#    -o --offset Offsets()
#       Better than the original, handles const(value) and compressed(size).
#    -s --slice  Slice()
#       Covered by test_show. Might as well keep this as a separate exe.
#    -a --alpha Alpha()
#       Not implemented, as the alpha tiles are deprecated.

def Main():
    np.seterr(all='raise')

    parser = argparse.ArgumentParser(description='Show ZGY file details')
    parser.add_argument('files', nargs="+", help='ZGY files, local or sd://')
    parser.add_argument('--histogram', action='store_true', help="Show the 256-bin histogram")
    parser.add_argument('--offsets', action='store_true', help="Show offset of each brick")
    parser.add_argument('--sorted-offsets', action='store_true', help="Sort by file offset")
    parser.add_argument('--only-lod0-info', action='store_true', help="Only lod0 stats etc.")
    args = parser.parse_args()
    for filename in args.files:
        run(filename, args)

if __name__ == "__main__":
    Main()

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



from numpy.core.fromnumeric import trace
import xarray as xr
import segyio
import numpy as np
import pandas as pd
import warnings

from ._segy_core import tqdm, check_tracefield

from .._accessor import open_seisnc
from .._core import FrozenDict
from .._keyfield import CoordKeyField
from ._segy_text import create_default_texthead, put_segy_texthead, _clean_texthead
from ._segy_globals import _ISEGY_MEASUREMENT_SYSTEM

TQDM_ARGS = dict(unit=" traces", desc="Writing to SEG-Y")

OUTPUT_BYTES = FrozenDict(
    dict(
        standard_3d=dict(iline=181, xline=185, cdp_x=189, cdp_y=193),
        standard_3d_gath=dict(iline=181, xline=185, cdp_x=189, cdp_y=193, offset=37),
        standard_2d=dict(cdp=21, cdp_x=189, cdp_y=193),
        standard_2d_gath=dict(cdp=21, cdp_x=189, cdp_y=193, offset=37),
        petrel_3d=dict(iline=5, xline=21, cdp_x=73, cdp_y=77),
        petrel_2d=dict(cdp=21, cdp_x=73, cdp_y=77),
    )
)


def ncdf2segy(
    ncfile,
    segyfile,
    CMP=False,
    iline=189,
    xline=193,
    il_chunks=10,
    dimension=None,
    silent=False,
):
    """Convert siesnc format (NetCDF4) to SEGY.

    Args:
        ncfile (string): The input SEISNC file
        segyfile (string): The output SEG-Y file
        CMP (bool, optional): The data is 2D. Defaults to False.
        iline (int, optional): Inline byte location. Defaults to 189.
        xline (int, optional): Crossline byte location. Defaults to 193.
        il_chunks (int, optional): The size of data to work on - if you have memory
            limitations. Defaults to 10.
        dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever is present
        silent (bool, optional): Turn off progress reporting. Defaults to False.
    """

    with open_seisnc(ncfile, chunks={"iline": il_chunks}) as seisnc:
        if dimension is None:
            if seisnc.seis.is_twt():
                dimension = "twt"
            elif seisnc.seis.is_depth():
                dimension = "depth"
            else:
                raise RuntimeError(
                    f"twt and depth dimensions missing, please specify a dimension to convert: {seisnc.dims}"
                )
        elif dimension not in seisnc.dims:
            raise RuntimeError(
                f"Requested output dimension not found in the dataset: {seisnc.dims}"
            )

        z0 = int(seisnc[dimension].values[0])
        ni, nj, nk = seisnc.dims["iline"], seisnc.dims["xline"], seisnc.dims[dimension]
        msys = _ISEGY_MEASUREMENT_SYSTEM[seisnc.seis.get_measurement_system()]
        spec = segyio.spec()

        # to create a file from nothing, we need to tell segyio about the structure of
        # the file, i.e. its inline numbers, crossline numbers, etc. You can also add
        # more structural information, but offsets etc. have sensible defautls. This is
        # the absolute minimal specification for a N-by-M volume
        spec.sorting = 1
        spec.format = 1
        spec.iline = iline
        spec.xline = xline
        spec.samples = range(nk)
        spec.ilines = range(ni)
        spec.xlines = range(nj)

        xl_val = seisnc["xline"].values
        il_val = seisnc["iline"].values

        coord_scalar = seisnc.coord_scalar
        if coord_scalar is None:
            coord_scalar = 0
        coord_scalar_mult = np.power(abs(coord_scalar), np.sign(coord_scalar) * -1)

        il_bags = _bag_slices(seisnc["iline"].values, n=il_chunks)

        with segyio.create(segyfile, spec) as segyf:
            for ilb in tqdm(il_bags, desc="Writing to SEGY", disable=silent):
                ilbl = range(ilb.start, ilb.stop, ilb.step)
                data = seisnc.isel(iline=ilbl)
                for i, il in enumerate(ilbl):
                    il0, iln = il * nj, (il + 1) * nj
                    segyf.header[il0:iln] = [
                        {
                            segyio.su.offset: 1,
                            iline: il_val[il],
                            xline: xln,
                            segyio.su.cdpx: int(cdpx * coord_scalar_mult),
                            segyio.su.cdpy: int(cdpy * coord_scalar_mult),
                            segyio.su.ns: nk,
                            segyio.su.delrt: z0,
                            segyio.su.scalco: int(coord_scalar),
                        }
                        for xln, cdpx, cdpy in zip(
                            xl_val, data.cdp_x.values[i, :], data.cdp_y.values[i, :]
                        )
                    ]
                    segyf.trace[il * nj : (il + 1) * nj] = data.data[
                        i, :, :
                    ].values.astype(np.float32)
            segyf.bin.update(
                tsort=segyio.TraceSortingFormat.INLINE_SORTING,
                hdt=int(seisnc.sample_rate * 1000),
                hns=nk,
                mfeet=msys,
                jobid=1,
                lino=1,
                reno=1,
                ntrpr=ni * nj,
                nart=ni * nj,
                fold=1,
            )


def output_byte_locs(name):
    """Return common bytes position variable_dict for segy_writer.

    Args:
        name (str): One of [standard_3d, petrel_3d]

    Returns:
        dict: A dictionary of SEG-Y byte positions and default seisnc variable
            pairs.

    Example:
        >>> segywriter(ncfile, 'segyfile.sgy', trace_header_map=output_byte_loc('petrel'))
        >>>
    """
    try:
        return OUTPUT_BYTES[name]
    except KeyError:
        raise ValueError(
            f"No byte locatons for {name}, select from {list(OUTPUT_BYTES.keys())}"
        )


def _bag_slices(ind, n=10):
    """Take a list of indices and create a list of bagged indices. Each bag
    will contain n indices except for the last bag which will contain the
    remainder.

    This function is designed to support bagging/chunking of data to support
    memory management or distributed processing.

    Args:
        ind (list/array-like): The input list to create bags from.
        n (int, optional): The number of indices per bag. Defaults to 10.

    Returns:
        [type]: [description]
    """
    if n is None:  # one bad
        return [slice(0, len(ind), 1)]
    else:
        bag = list()
        prev = 0
        i = 0
        for i in range(len(ind)):
            if (i + 1) % n == 0:
                bag.append(slice(prev, i + 1, 1))
                prev = i + 1
        if prev != i + 1:
            bag.append(slice(prev, i + 1, 1))
        return bag


def _check_dimension(seisnc, dimension):
    if dimension is None:
        if seisnc.seis.is_twt():
            dimension = "twt"
        elif seisnc.seis.is_depth():
            dimension = "depth"
        else:
            raise RuntimeError(
                f"twt and depth dimensions missing, please specify a dimension to convert: {seisnc.dims}"
            )
    elif dimension not in seisnc.dims:
        raise RuntimeError(
            f"Requested output dimension not found in the dataset: {seisnc.dims}"
        )

    return dimension


def _build_trace_headers(ds, vert_dimension, other_dim_order, thm_args, cdp_x, cdp_y):
    # build trace headers for iteration and output
    trace_layout = ds.data.isel(**{vert_dimension: 0})
    trace_headers = pd.DataFrame()
    for var in thm_args:
        try:
            trace_headers[thm_args[var]] = (
                ds[var]
                .broadcast_like(trace_layout)
                .transpose(*other_dim_order, transpose_coords=True)
                .values.ravel()
            )
        except KeyError:  # The requested variable isn't in the list, write zeros
            warnings.warn(
                f"The variable {var} was not found in the input data, writing zeros to header instead"
            )

            if ds.seis.is_2d() or ds.seis.is_2dgath():
                base_var = "cdp"
            else:
                base_var = "iline"

            trace_headers[thm_args[var]] = (
                ds[base_var]
                .broadcast_like(trace_layout)
                .transpose(*other_dim_order, transpose_coords=True)
                .values.ravel()
                * 0.0
            )
    # scale coords
    trace_headers[cdp_x] = trace_headers[cdp_x] * ds.coord_scalar_mult
    trace_headers[cdp_y] = trace_headers[cdp_y] * ds.coord_scalar_mult
    trace_headers[cdp_x] = trace_headers[cdp_x]
    trace_headers[cdp_y] = trace_headers[cdp_y]

    trace_headers = trace_headers.astype(np.int32)

    return trace_headers


def _ncdf2segy_3d(
    ds,
    segyfile,
    iline=None,
    xline=None,
    cdp_x=None,
    cdp_y=None,
    il_chunks=10,
    dimension=None,
    text=None,
    silent=False,
    **thm_args,
):
    """Convert siesnc format (NetCDF4) to SEGY.

    Args:
        ds (xarray.Dataset): The input SEISNC dataset
        segyfile (string): The output SEG-Y file
        iline (int): Inline byte location.
        xline (int): Crossline byte location.
        cdp_x (int): The byte location to write the cdp_x.
        cdp_y (int): The byte location to write the cdp_y.
        il_chunks (int, optional): The size of data to work on - if you have memory
            limitations. Defaults to 10.
        dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever is present
        text (dict, optional): A dictionary of strings to write out to the text header.
        silent (bool, optional): Turn off progress reporting. Defaults to False.
        thm_args (int): Byte locations to write trace header variables defined by key word arguments.
            keys for thm_args must be variables or coordinates or dimensions in seisnc.
    """

    thm_args[CoordKeyField.iline] = iline
    thm_args[CoordKeyField.xline] = xline
    thm_args[CoordKeyField.cdp_x] = cdp_x
    thm_args[CoordKeyField.cdp_y] = cdp_y

    z0 = int(ds[dimension].values[0])
    ni, nj, nk = (
        ds.dims[CoordKeyField.iline],
        ds.dims[CoordKeyField.xline],
        ds.dims[dimension],
    )
    order = (CoordKeyField.iline, CoordKeyField.xline)
    msys = _ISEGY_MEASUREMENT_SYSTEM[ds.seis.get_measurement_system()]
    spec = segyio.spec()

    # to create a file from nothing, we need to tell segyio about the structure of
    # the file, i.e. its inline numbers, crossline numbers, etc. You can also add
    # more structural information, but offsets etc. have sensible defautls. This is
    # the absolute minimal specification for a N-by-M volume
    spec.sorting = 2
    spec.format = 1
    spec.iline = iline
    spec.xline = xline
    spec.samples = range(nk)
    spec.ilines = range(ni)
    spec.xlines = range(nj)

    trace_headers = _build_trace_headers(ds, dimension, order, thm_args, cdp_x, cdp_y)

    # header constants
    for byte, val in zip(
        [segyio.su.offset, segyio.su.ns, segyio.su.delrt, segyio.su.scalco],
        [1, nk, z0, int(ds.coord_scalar)],
    ):
        if byte not in trace_headers.columns:  # don't override user values
            trace_headers[byte] = val

    # to records
    trace_headers = trace_headers.to_dict(orient="records")

    il_bags = _bag_slices(ds[CoordKeyField.iline].values, n=il_chunks)
    with segyio.create(segyfile, spec) as segyf:
        pbar = tqdm(total=ni * nj, disable=silent, **TQDM_ARGS)
        for ilb in il_bags:
            ilbl = range(ilb.start, ilb.stop, ilb.step)
            data = ds.isel(iline=ilbl).transpose(*order, dimension)
            for i, il in enumerate(ilbl):
                il0, iln = il * nj, (il + 1) * nj
                segyf.header[il0:iln] = trace_headers[il0:iln]
                segyf.trace[il0:iln] = data.data[i, :, :].values.astype(np.float32)
                pbar.update(nj)
        pbar.close()

    if not text:
        # create text header
        overrides = {
            #     123456789012345678901234567890123456789012345678901234567890123456
            7: f"Original File: {ds.source_file}",
            35: "*** BYTE LOCATION OF KEY HEADERS ***",
            36: f"CMP UTM-X: {cdp_x}, ALL COORDS SCALED BY: {ds.coord_scalar_mult},"
            f" CMP UTM-Y: {cdp_y}",
            37: f"INLINE: {iline}, XLINE: {xline}, ",
            40: "END TEXTUAL HEADER",
        }
        put_segy_texthead(
            segyfile,
            create_default_texthead(overrides),
        )
    else:
        put_segy_texthead(segyfile, text, line_counter=False)


def _ncdf2segy_3dgath(
    ds,
    segyfile,
    iline=None,
    xline=None,
    offset=None,
    cdp_x=None,
    cdp_y=None,
    il_chunks=10,
    dimension=None,
    text=None,
    silent=False,
    **thm_args,
):
    """Convert siesnc format (NetCDF4) to SEGY.

    Args:
        ds (xarray.Dataset): The input SEISNC dataset
        segyfile (string): The output SEG-Y file
        iline (int): Inline byte location.
        xline (int): Crossline byte location.
        cdp_x (int): The byte location to write the cdp_x.
        cdp_y (int): The byte location to write the cdp_y.
        offset (int): Offset byte location.
        il_chunks (int, optional): The size of data to work on - if you have memory
            limitations. Defaults to 10.
        dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever is present
        text (dict, optional): A dictionary of strings to write out to the text header.
        silent (bool, optional): Turn off progress reporting. Defaults to False.
        thm_args (int): Byte locations to write trace header variables defined by key word arguments.
            keys for thm_args must be variables or coordinates or dimensions in seisnc.
    """

    thm_args[CoordKeyField.iline] = iline
    thm_args[CoordKeyField.xline] = xline
    thm_args[CoordKeyField.cdp_x] = cdp_x
    thm_args[CoordKeyField.cdp_y] = cdp_y
    thm_args[CoordKeyField.offset] = offset

    z0 = int(ds[dimension].values[0])
    ni, nj, nk, nl = (
        ds.dims[CoordKeyField.iline],
        ds.dims[CoordKeyField.xline],
        ds.dims[dimension],
        ds.dims[CoordKeyField.offset],
    )
    order = (CoordKeyField.iline, CoordKeyField.xline, CoordKeyField.offset)
    msys = _ISEGY_MEASUREMENT_SYSTEM[ds.seis.get_measurement_system()]
    spec = segyio.spec()

    # to create a file from nothing, we need to tell segyio about the structure of
    # the file, i.e. its inline numbers, crossline numbers, etc. You can also add
    # more structural information, but offsets etc. have sensible defautls. This is
    # the absolute minimal specification for a N-by-M volume
    spec.format = 2
    spec.iline = iline
    spec.xline = xline
    spec.samples = ds[dimension].astype(int).values
    spec.ilines = ds[CoordKeyField.iline].astype(int).values
    spec.xlines = ds[CoordKeyField.xline].astype(int).values
    spec.offsets = ds[CoordKeyField.offset].astype(int).values

    trace_headers = _build_trace_headers(ds, dimension, order, thm_args, cdp_x, cdp_y)

    # header constants
    for byte, val in zip(
        [segyio.su.ns, segyio.su.delrt, segyio.su.scalco],
        [nk, z0, int(ds.coord_scalar)],
    ):
        if byte not in trace_headers.columns:  # don't override user values
            trace_headers[byte] = val

    # to records
    trace_headers = trace_headers.to_dict(orient="records")

    # print(len(trace_headers))

    il_bags = _bag_slices(ds[CoordKeyField.iline].values, n=il_chunks)
    with segyio.create(segyfile, spec) as segyf:
        pbar = tqdm(total=nj * nl * ni, disable=silent, **TQDM_ARGS)
        for ilb in il_bags:
            ilbl = range(ilb.start, ilb.stop, ilb.step)
            data = ds.isel(iline=ilbl).transpose(*order, dimension)
            for i, il in enumerate(ilbl):
                t0, tn = il * nj * nl, (il + 1) * nj * nl
                segyf.header[t0:tn] = trace_headers[t0:tn]
                segyf.trace[t0:tn] = (
                    data.data[i, :, :, :].values.reshape(nj * nl, nk).astype(np.float32)
                )
                pbar.update(nj * nl)
        pbar.close()
        segyf.bin.update(
            tsort=segyio.TraceSortingFormat.UNKNOWN_SORTING,  # trace sorting
            hdt=int(ds.sample_rate * 1000),  # Interval
            hns=nk,  # samples
            mfeet=msys,  # measurement system
            jobid=1,  # jobid
            lino=1,  # line number
            reno=1,  # reel number
            ntrpr=ni * nj * nl,  # n traces
            nart=0,  # aux traces
            fold=1,
        )

    if not text:
        # create text header
        overrides = {
            #     123456789012345678901234567890123456789012345678901234567890123456
            7: f"Original File: {ds.source_file}",
            35: "*** BYTE LOCATION OF KEY HEADERS ***",
            36: f"CMP UTM-X: {cdp_x}, ALL COORDS SCALED BY: {ds.coord_scalar_mult},"
            f" CMP UTM-Y: {cdp_y}",
            37: f"INLINE: {iline}, XLINE: {xline}, ",
            40: "END TEXTUAL HEADER",
        }
        put_segy_texthead(segyfile, create_default_texthead(overrides))
    else:
        put_segy_texthead(segyfile, text, line_counter=False)


def _ncdf2segy_2d(
    ds,
    segyfile,
    cdp=None,
    cdp_x=None,
    cdp_y=None,
    dimension=None,
    text=None,
    silent=False,
    **thm_args,
):
    """Convert siesnc format (NetCDF4) to SEGY.

    Args:
        ds (xarray.Dataset): The input SEISNC dataset
        segyfile (string): The output SEG-Y file
        cdp (int): The byte location to write the cdp.
        cdp_x (int): The byte location to write the cdp_x.
        cdp_y (int): The byte location to write the cdp_y.
        dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever is present
        text (dict, optional): A dictionary of strings to write out to the text header.
        silent (bool, optional): Turn off progress reporting. Defaults to False.
        thm_args (int): Byte locations to write trace header variables defined by key word arguments.
            keys for thm_args must be variables or coordinates or dimensions in seisnc.
    """

    thm_args[CoordKeyField.cdp] = cdp
    thm_args[CoordKeyField.cdp_x] = cdp_x
    thm_args[CoordKeyField.cdp_y] = cdp_y

    z0 = int(ds[dimension].values[0])
    ni, nk = (
        ds.dims[CoordKeyField.cdp],
        ds.dims[dimension],
    )
    order = (CoordKeyField.cdp,)
    msys = _ISEGY_MEASUREMENT_SYSTEM[ds.seis.get_measurement_system()]
    spec = segyio.spec()

    # to create a file from nothing, we need to tell segyio about the structure of
    # the file, i.e. its inline numbers, crossline numbers, etc. You can also add
    # more structural information, but offsets etc. have sensible defautls. This is
    # the absolute minimal specification for a N-by-M volume
    spec.format = 1
    spec.iline = cdp
    spec.xline = 185
    spec.samples = range(nk)
    spec.tracecount = ni

    trace_headers = _build_trace_headers(ds, dimension, order, thm_args, cdp_x, cdp_y)

    # header constants
    for byte, val in zip(
        [segyio.su.offset, segyio.su.ns, segyio.su.delrt, segyio.su.scalco],
        [1, nk, z0, int(ds.coord_scalar)],
    ):
        if byte not in trace_headers.columns:  # don't override user values
            trace_headers[byte] = val

    # to records
    trace_headers = trace_headers.to_dict(orient="records")

    with segyio.create(segyfile, spec) as segyf:
        data = ds.transpose(*order, dimension)
        segyf.header[:] = trace_headers
        segyf.trace[:] = data.data.values.astype(np.float32)

        segyf.bin.update(
            tsort=segyio.TraceSortingFormat.INLINE_SORTING,
            hdt=int(ds.sample_rate * 1000),
            hns=nk,
            mfeet=msys,
            jobid=1,
            lino=1,
            reno=1,
            ntrpr=ni,
            nart=0,
            fold=1,
        )

    if not text:
        # create text header
        overrides = {
            #     123456789012345678901234567890123456789012345678901234567890123456
            7: f"Original File: {ds.source_file}",
            35: "*** BYTE LOCATION OF KEY HEADERS ***",
            36: f"CMP UTM-X: {cdp_x}, ALL COORDS SCALED BY: {ds.coord_scalar_mult},"
            f" CMP UTM-Y: {cdp_y}",
            37: f"CDP: {cdp}, ",
            40: "END TEXTUAL HEADER",
        }
        put_segy_texthead(
            segyfile,
            create_default_texthead(overrides),
        )
    else:
        put_segy_texthead(segyfile, text, line_counter=False)


def _ncdf2segy_2d_gath(
    ds,
    segyfile,
    cdp=None,
    cdp_x=None,
    cdp_y=None,
    offset=None,
    dimension=None,
    text=None,
    silent=False,
    **thm_args,
):
    """Convert siesnc format (NetCDF4) to SEGY.

    Args:
        ds (xarray.Dataset): The input SEISNC dataset
        segyfile (string): The output SEG-Y file
        cdp (int): The byte location to write the cdp.
        cdp_x (int): The byte location to write the cdp_x.
        cdp_y (int): The byte location to write the cdp_y.
        offset (int): The byte location to write the offset.
        dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever is present
        text (dict, optional): A dictionary of strings to write out to the text header.
        silent (bool, optional): Turn off progress reporting. Defaults to False.
        thm_args (int): Byte locations to write trace header variables defined by key word arguments.
            keys for thm_args must be variables or coordinates or dimensions in seisnc.
    """

    thm_args[CoordKeyField.cdp] = cdp
    thm_args[CoordKeyField.cdp_x] = cdp_x
    thm_args[CoordKeyField.cdp_y] = cdp_y
    thm_args[CoordKeyField.offset] = offset

    z0 = int(ds[dimension].values[0])
    ni, nk, nl = (
        ds.dims[CoordKeyField.cdp],
        ds.dims[dimension],
        ds.dims[CoordKeyField.offset],
    )
    order = (CoordKeyField.cdp, CoordKeyField.offset)
    msys = _ISEGY_MEASUREMENT_SYSTEM[ds.seis.get_measurement_system()]
    spec = segyio.spec()

    # to create a file from nothing, we need to tell segyio about the structure of
    # the file, i.e. its inline numbers, crossline numbers, etc. You can also add
    # more structural information, but offsets etc. have sensible defautls. This is
    # the absolute minimal specification for a N-by-M volume
    spec.format = 1
    spec.iline = cdp
    spec.xline = 185
    spec.samples = range(nk)
    spec.tracecount = ni * nl

    trace_headers = _build_trace_headers(ds, dimension, order, thm_args, cdp_x, cdp_y)

    # header constants
    for byte, val in zip(
        [segyio.su.ns, segyio.su.delrt, segyio.su.scalco],
        [nk, z0, int(ds.coord_scalar)],
    ):
        if byte not in trace_headers.columns:  # don't override user values
            trace_headers[byte] = val

    # to records
    trace_headers = trace_headers.to_dict(orient="records")

    with segyio.create(segyfile, spec) as segyf:
        data = ds.transpose(*order, dimension)
        segyf.header[:] = trace_headers
        segyf.trace[:] = data.data.values.reshape(-1, nk).astype(np.float32)

        segyf.bin.update(
            tsort=segyio.TraceSortingFormat.INLINE_SORTING,
            hdt=int(ds.sample_rate * 1000),
            hns=nk,
            mfeet=msys,
            jobid=1,
            lino=1,
            reno=1,
            ntrpr=ni * nl,
            nart=0,
            fold=1,
        )

    if not text:
        # create text header
        overrides = {
            #     123456789012345678901234567890123456789012345678901234567890123456
            7: f"Original File: {ds.source_file}",
            35: "*** BYTE LOCATION OF KEY HEADERS ***",
            36: f"CMP UTM-X: {cdp_x}, ALL COORDS SCALED BY: {ds.coord_scalar_mult},"
            f" CMP UTM-Y: {cdp_y}",
            37: f"CDP: {cdp}, OFFSET: {offset}",
            40: "END TEXTUAL HEADER",
        }
        put_segy_texthead(
            segyfile,
            create_default_texthead(overrides),
        )
    else:
        put_segy_texthead(segyfile, text, line_counter=False)


def _segy_writer_input_handler(
    ds, segyfile, trace_header_map, dimension, silent, il_chunks, text
):
    """Handler for open seisnc dataset"""

    dimension = _check_dimension(ds, dimension)

    # ensure there is a coord_scalar
    coord_scalar = ds.coord_scalar
    if coord_scalar is None:
        coord_scalar = 0
        ds.attrs["coord_scalar"] = coord_scalar
    coord_scalar_mult = np.power(abs(coord_scalar), np.sign(coord_scalar) * -1)
    ds.attrs["coord_scalar_mult"] = coord_scalar_mult

    # create empty trace header map if necessary
    if trace_header_map is None:
        trace_header_map = dict()
    else:  # check that values of thm in ds
        for key in trace_header_map:
            try:
                _ = ds[key]
            except KeyError:
                raise ValueError("keys of trace_header_map must be in seisnc")

    # transfrom text if requested
    if text:
        text = {i + 1: line for i, line in enumerate(ds.text.split("\n"))}
        text = _clean_texthead(text, 80)

    common_kwargs = dict(silent=silent, dimension=dimension, text=text)

    if ds.seis.is_2d():
        thm = output_byte_locs("standard_2d")
        thm.update(trace_header_map)
        _ncdf2segy_2d(ds, segyfile, **common_kwargs, **thm)
    elif ds.seis.is_2dgath():
        thm = output_byte_locs("standard_2d_gath")
        thm.update(trace_header_map)
        _ncdf2segy_2d_gath(ds, segyfile, **common_kwargs, **thm)
    elif ds.seis.is_3d():
        thm = output_byte_locs("standard_3d")
        thm.update(trace_header_map)
        _ncdf2segy_3d(
            ds,
            segyfile,
            **common_kwargs,
            il_chunks=il_chunks,
            **thm,
        )
    elif ds.seis.is_3dgath():
        thm = output_byte_locs("standard_3d_gath")
        thm.update(trace_header_map)
        _ncdf2segy_3dgath(
            ds,
            segyfile,
            **common_kwargs,
            il_chunks=il_chunks,
            **thm,
        )
    else:
        # cannot determine type of data writing a continuous traces
        raise NotImplementedError()


def segy_writer(
    seisnc,
    segyfile,
    trace_header_map=None,
    il_chunks=None,
    dimension=None,
    silent=False,
    use_text=False,
):
    """Convert siesnc format (NetCDF4) to SEGY.

    Args:
        seisnc (xarray.Dataset, string): The input SEISNC file either a path or the in memory xarray.Dataset
        segyfile (string): The output SEG-Y file
        trace_header_map (dict, optional): Defaults to None. A dictionary of seisnc variables
            and byte locations. The variable will be written to the trace headers in the
            assigned byte location. By default CMP=23, cdp_x=181, cdp_y=185, iline=189,
            xline=193.
        il_chunks (int, optional): The size of data to work on - if you have memory
            limitations. Defaults to 10. This is primarily used for large 3D and ignored for 2D data.
        dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever is present
        silent (bool, optional): Turn off progress reporting. Defaults to False.
        use_text (bool, optional): Use the seisnc text for the EBCIDC output. This text usally comes from
            the loaded SEG-Y file and may not match the segysak SEG-Y output. Defaults to False and writes
            the default segysak EBCIDC
    """
    if trace_header_map:
        check_tracefield(trace_header_map.values())

    if isinstance(seisnc, xr.Dataset):
        _segy_writer_input_handler(
            seisnc, segyfile, trace_header_map, dimension, silent, None, use_text
        )
    else:
        ncfile = seisnc
        if isinstance(il_chunks, int):
            chunking = {"iline": il_chunks}
        else:
            chunking = None
        with open_seisnc(ncfile, chunks=chunking) as seisnc:
            _segy_writer_input_handler(
                seisnc,
                segyfile,
                trace_header_map,
                dimension,
                silent,
                il_chunks,
                use_text,
            )



#!/usr/bin/env python3

##@package openzgy.impl.compress

import math
import sys
import numpy as np
from collections import namedtuple

from ..impl import enum as impl_enum
from ..exception import *

CompressionResult = namedtuple("Compression", "cdata csize isize signal noise ctime dtime snr_result snr_wanted snr_rounded snr_step")
CompressionResult.__doc__ = """The result returned from _xxx_try_one() functions.
The first member, cdata, is the one that really matters.
The snr_xxx members may be useful for goal seek.
All members except cdata may be useful logging and testing.
"""
CompressionResult.cdata.__doc__ = "bytes-like compressed data"
CompressionResult.csize.__doc__ = "Size in bytes of the compressed data"
CompressionResult.isize.__doc__ = "Size in bytes of the original input data"
CompressionResult.signal.__doc__ = "Sum of all sample values in the input"
CompressionResult.noise.__doc__ = "Sum of all absolute errors in round trip"
CompressionResult.ctime.__doc__ = "Time in seconds used for compression"
CompressionResult.dtime.__doc__ = "Time in seconds used for decompression"
CompressionResult.snr_result.__doc__ = "The SNR that was achieved"
CompressionResult.snr_wanted.__doc__ = "The SNR that was requested by the user"
CompressionResult.snr_rounded.__doc__ = "The SNR that was requested from the algorithm"
CompressionResult.snr_result.__doc__ = "Distance to next or previous quality level"

class CompressPlugin:
    """
    Base class for OpenZGY compression plug-ins.
    If anybody wants to add additional compression algorithms it
    is recommended but not required to use this base class. See
    CompressFactoryImpl.register{Compressor,Decompressor} for how to
    use plain functors (C++) or callables (Python) instead.

    This class performs triple duty as it handles both compression
    and decompression static methods (need not have been together)
    and an instance of the class can be used a compressor functor
    if a lambda is too limiting. To invoke the methods:

        MyCompressPlugin.factory(...)(data)
        MyCompressPlugin.compress(data, ...) (NOT recommended)
        MyCompressPlugin.decompress(cdata,status,shape,file_dtype,user_dtype)

    The following will also work but should only be used for very simple
    compressors that have no parameters. In the first case MyCompressPlugin
    won't have the option to return None for certain parameters, and in the
    second case handling a variable arguent list becomes trickier.

    To register this class:
      CompressFactoryImpl.registerCompressor("My",MyCompressPlugin.factory)
      CompressFactoryImpl.registerDecompressor("My",MyCompressPlugin.decompress)

    To use the compression part from client code:
       compressor = ZgyCompressFactory("My", ...)
    """
    @staticmethod
    def compress(data, *args, **kwargs):
        """
        This is an abstract method.

        Compress a 3d or (TODO-Low 2d) numpy array, returning a bytes-like
        result. If called with a single "data" argument the compression
        will be done with default parameters and no extended logging.

        Additional arguments are specific to the compression type.

        The function can be used directly as the compression hook.
        But you probably want a lambda expression or a real instance
        of this class instead, to be able to specify parameters.

        The compression algorithm is used is assumed to handle big / little
        endian conversion itself. TODO-Worry this is not quite true for ZFP.
        See the documentation. A special compilation flag is needed
        on big endian machines. Also I suspect the optional hedaer
        (which this code uses) might need byte swapping.
        """
        raise ZgyInternalError("Attempt to invoke abstract method")

    @staticmethod
    def decompress(cdata, status, shape, file_dtype, user_dtype):
        """
        This is an abstract method.

        Decompress bytes or similar into a numpy.ndarray.

        Arguments:
          cdata      -- bytes or bytes-like compressed data,
                        possibly with trailing garbage.
          status     -- Currently always BrickStatus.Compressed,
                        in the future the status might be used to
                        distinguish between different compression
                        algorithms instead of relying on magic numbers.
          shape      -- Rank and size of the result in case this is
                        not encoded by the compression algorithm.
          file_dtype -- Original value type before compression,
                        in case the decompressor cannot figure it out.
                        This will exactly match the dtype of the
                        data buffer passed to the compressor.
          user_dtype -- Required value type of returned array.

        Passing an uncompressed brick to this function is an error.
        We don't have enough context to handle uncompressed bricks
        that might require byteswapping and fix for legacy quirks.
        Also cannot handle constant bricks, missing bricks, etc.

        The reason user_dtype is needed is to avoid additional
        quantization noise when the user requests integer compressed data
        to be read as float. the decompressor might need to convert
        float data to int, only to have it converted back to float later.

        Current assumptions made of all candidate algorithms:

            -  The compressed data stream may have trailing garbage;
               this will be silently ignored by the decompressor.

            -  The compressed data stream will never be longer than
               the uncompressed data. This needs to be enforced by
               the compressor. The compressor is allowed to give up
               and tell the caller to not compress this brick.

            -  The reason for the two assumptions above is an
               implementation detail; the reported size of a
               compressed brick is not completely reliable.
               This might change in the next version

            -  The compressed data stream must start with a magic
               number so the decompressor can figure out whether
               this is the correct algorithm to use.

        If the assumptions cannot be met, the compressor / decompressor
        for this particular type could be modified to add an extra header
        with the compressed size and a magic number. Or we might add a
        (size, algorithm number) header to every compressed block to
        relieve the specific compressor / decompressor from worrying
        about this. Or the brick status could be used to encode which
        algorithm was used, picked up from the MSB of the lup entry.
        Which would also require the compressor to return both the
        actual compressed data and the code to identify the decompressor.
        That is the main reason we are also passed the "status" arg.
        Caveat: If adding an extra header, keep in mind that this header
        must be included when checking that the compressed stream is not
        too big.
        """
        raise ZgyInternalError("Attempt to invoke abstract method")

    def __init__(self, *args, **kwargs):
        """
        Create an instance that remembers the arguments it was created with.
        When the instance is called as a function it will invoke compress()
        with those arguments. So you can use either of the following:

            compressor = CompressPlugin.compress # no arguments
            compressor = CompressPlugin(...)
            compressor = lambda x: CompressPlugin.compress(x, ...)

        Derived classes don't need to redefine __init__ and __call__.
        But they might want to in order to get argument checking.
        The __init__ in the base class accepts any arguments so an error
        won't be caught until the first time the compressor is invoked.
        """
        self._args = args
        self._kwargs = kwargs

    def __call__(self, data):
        """
        Invoke the compressor with arguments passed by the constructor.
        """
        return self.compress(data, *args, **kwargs)

    def dump(*args, **kwargs):
        """
        Output statistics to standard output, if possible.
        """
        pass

class CompressFactoryImpl:
    """
    Registry of known compress and decompress algorithms.
    Those two are completely separate but we might as well
    handle both in the same class.
    """
    _compress_registry = {}
    _decompress_registry = []

    @staticmethod
    def registerCompressor(name, fn):
        """
        Register a factory function that will be called to create
        a function that in turn can be used to compress a data block.
        Pass fn = None if you for some reason need to remove a registration.

        The registered function can have any signature; the signature
        needs to include whatever parameters the actual compressor wants.

        The function that is created by the factory must have the signature:
            raw:          bytes.
            brickstatus:  impl.enum.BrickStatus,
            bricksize:    tuple(int,int,int),
            file_dtype:   np.dtype,
            result_dtype: np.dtype

        The function's return value:
            np.ndarray with rank 3, shape bricksize, and dtype result_dtype.

        Example usage of the factory:
            old: with ZgyWriter(snr=snr)
            new: with ZgyWriter(compressor = ZgyCompressFactory("ZFP", snr=30),

        The example shows that this is a bit more inconvenient for the end
        user. But it allows for using different compression plug-ins with
        arbitrary parameters.

        Note that user code doesn't need to use ZgyCompressFactory() and
        its list of known compressors. Instead a compression function
        can be provided directly. But most likely the factory method will
        be simpler to maintain.

        fn() is allowed to return None, which will have the same effect
        as if the user did not specify any compression. E.g. there might be
        a convention that snr<0 means store uncompressed. Allowing the
        factory to return None in that case means the example above would
        still work. Otherwise the snr<0 test would be included in the
        client code. Which is messy.
        """
        CompressFactoryImpl._compress_registry[name] = fn

    @staticmethod
    def registerDecompressor(name, fn):
        """
        Register a factory function that is able to decompress one or more
        types of compressed data. The registered functions will be called
        in reverse order of registration until one of them indicates that
        it has decompressed the data. You cannot remove a registration
        but you can effectively disable it by registering another one
        that recognizes the same input data.
        The supplied name is only for information.

        The function that is created by the factory must have the signature:
            raw:          bytes.
            brickstatus:  impl.enum.BrickStatus,
            bricksize:    tuple(int,int,int),
            file_dtype:   np.dtype,
            result_dtype: np.dtype
        """
        CompressFactoryImpl._decompress_registry.insert(0, (name, fn))

    def knownCompressors():
        """
        Return the names of all compressors known to the system.
        This is primarily for logging, but might in principle be used
        in a GUI to present a list of compressors to choose from.
        The problem with that is how to handle the argument list.
        """
        return list([k for k, v in CompressFactoryImpl._compress_registry.items()])

    def knownDecompressors():
        """
        Return the names of all compressors known to the system.
        This is primarily for logging.
        """
        return list([k for k, v in CompressFactoryImpl._decompress_registry])

    @staticmethod
    def factory(name, *args, **kwargs):
        fn = CompressFactoryImpl._compress_registry.get(name, None)
        if not fn:
            known = ",".join(CompressFactoryImpl.knownDecompressors())
            raise ZgyMissingFeature('Compression algorithm "{0}" not recognized. Must be one of ({1}).'.format(name, known))
        return fn(*args, **kwargs)

    @staticmethod
    def decompress(cdata, status, shape, file_dtype, user_dtype):
        """
        Loop over all registered decompressors and try to find one that
        can handle this particular brick. Raises an error if none found.
        See CompressPlugin.decompress() for parameter descriptions.
        """
        result = None
        if status != impl_enum.BrickStatus.Compressed:
            raise ZgyInternalError("Tried to decompress uncompressed data.")
        for k, v in CompressFactoryImpl._decompress_registry:
            result = v(cdata, status, shape, file_dtype, user_dtype)
            if result is not None:
                break
        if result is None:
            raise ZgyFormatError("Compression algorithm not recognized.")
        elif tuple(result.shape) != tuple(shape):
            raise ZgyFormatError("Decompression returned unexpected data.")
        return result

class CompressStats:
    def __init__(self, details):
        self._details  = details # Parameters etc, just for logging.
        self._signal   = 0
        self._noise    = 0
        self._original = 0
        self._packed   = 0
        self._types    = dict()
        self._lossy    = 0
        self._perfect  = 0
        self._all_info = [] # (measured_snr, measured_compressed_size_%)
        self._timed_bytes = 0
        self._ctime = 0
        self._dtime = 0

    def add_data(self, idata, csize, ddata, *, ctime = None, dtime = None, msg = None):
        signal, noise = self._collect_snr(idata, ddata)
        isize = idata.size * idata.itemsize
        snr = self._compute_snr(signal, noise)
        self.add(signal=signal, noise=noise, snr=snr, isize=isize, csize=csize,
                 ctime=ctime, dtime=dtime, msg=msg)
        #print("signal {0:8.0f} noise {1:8.0f} snr {2:.2f}".format(signal, noise, snr))

    def add(self, signal, noise, snr, isize, csize, *, ctime = None, dtime = None, msg = None):
        self._all_info.append([snr, 100*csize/isize])
        if snr < 99:
            # s/n only logged for lossy compression.
            self._lossy    += 1
            self._signal   += signal
            self._noise    += noise
        else:
            self._perfect  += 1
        self._original += isize
        self._packed   += csize
        if ctime is not None and dtime is not None:
            self._timed_bytes += isize
            self._ctime += ctime
            self._dtime += dtime
        if msg == "compress" and snr >= 99: msg = "compres*"
        if msg:
            self._types[msg] = 1 + self._types.setdefault(msg, 0)
        if False and msg:
            print("@@ yoo-hoo {0} {1} -> {2} factor {3:.1f} snr {4:.1f}".format(
                msg, isize, csize, isize/csize if csize else np.nan, snr))

    @staticmethod
    def _collect_snr(idata, ddata):
        """
        This function along with _compute_snr defines the cost function
        used when trying to quantify noise. The input data should not
        contain NaN or Inf. We won't be applying lossy compression to
        that kind of data anyway.

        There is no perfect solution. We want to avoid having
        neighboring bricks showning significant differences in
        quality. This sounds easy enough, but "similar quality"
        depends very much on the cost function.
        """
        # Computationally inexpensive, but has problems with spikes.
        # A brick containing many spikes will appear to have a higher
        # average amplitude, which allows the compressor to leave
        # more noise in it. It would have been better to use the
        # average signal level of the entire survey instead of the
        # brick. But that information is not available yet.
        signal = np.sum(np.abs(idata), dtype=np.float64)
        noise  = np.sum(np.abs(idata - ddata), dtype=np.float64)
        # Expensive, and has problems with dead traces and bricks
        # that contain more than 50% water. The signal and noise
        # in thise cases would be measuring only the dead and/or
        # water samples.
        #signal = np.median(np.abs(idata)) * idata.size
        #noise  = np.median(np.abs(idata - ddata)) * idata.size
        return signal, noise

    @staticmethod
    def _compute_snr(s, n):
        return 99 if n == 0 else -10 if s/n < 0.315 else min(99, 6 * math.log2(s/n))

    def snr(self):
        return self._compute_snr(self._signal, self._noise)

    def dump(self, msg = None, *, outfile=None, text=True, csv=False):
        outfile = outfile or sys.stdout
        # Measured SNR for each brick, excluding lossless unless everything
        # is lossless. In which case we pretend there is just one brick.
        # The reason is that the median compression makes more sense
        # that way. If I had been plotting all percentiles instead of
        # just printing the median it would have made more sense to include
        # the lossless (SNR 99) bricks as well.
        all_snr = list([x[0] for x in self._all_info if x[0] < 99]) or [99.0]
        all_snr.sort()
        # Measured compression ratio for each brick, excluding uncompressed
        # unless everything is uncompressed. In which case we pretend there
        # is just one brick. Note that lossless can still be compressed.
        ratios = list([ 100.0/e[1] for e in self._all_info if e[1] != 1 and e[1] != 0 ]) or [1.0]
        ratios.sort()
        good = self._original and self._packed
        args = {
            "msg":          msg or "Processed",
            "details":      self._details,
            "orig_mb":      self._original / (1024*1024),
            "pack_mb":      self._packed,
            "factor":       self._original / self._packed if good else -1,
            "percent":      100 * self._packed / self._original if good else 1,
            "n_perfect":    self._perfect,
            "n_lossy":      self._lossy,
            "snr":          self.snr(),
            "snr_min":      all_snr[0],
            "snr_med":      all_snr[len(all_snr)//2],
            "snr_max":      all_snr[-1],
            "factor_min":   ratios[0],
            "factor_med":   ratios[len(ratios)//2],
            "factor_max":   ratios[-1],
            "percent_max":  100 / ratios[0],
            "percent_med":  100 / ratios[len(ratios)//2],
            "percent_min":  100 / ratios[-1],
            "ctime":        0 if not self._ctime else (self._timed_bytes / self._ctime) / (1024*1024),
            "dtime":        0 if not self._dtime else (self._timed_bytes / self._dtime) / (1024*1024),
            "_ctime": self._ctime,
            "_dtime": self._dtime,
            "_cplusdtime": self._ctime + self._dtime,
            "_timed_bytes": self._timed_bytes / (1024*1024),
        }
        if text:
            print("{msg} {orig_mb:.0f} MB, compression factor {factor:.1f} (median {factor_med:.1f} with {n_perfect} bricks lossless and {n_lossy} bricks min snr {snr_min:.1f} median {snr_med:.1f} overall {snr:.1f}) compress {ctime:.1f} MB/s decompress {dtime:.1f} MB/s".format(**args), file=outfile)
            for k in sorted(self._types):
                print("   ", k, self._types[k], file=outfile)
        if csv:
            if csv == "header":
                print(";msg;requested snr;median snr;snr;compressed size%;median size%;compress MB/s;decompress MB/s;compress elapsed;decompress elapsed;sum elapsed;Total data MB", file=outfile)
            else:
                print(";{msg};{details};{snr_med:.1f};{snr:.1f};{percent:.1f};{percent_med:.1f};{ctime:.1f};{dtime:.1f};{_ctime:.1f};{_dtime:.1f};{_cplusdtime:.1f};{_timed_bytes:.1f}".format(**args), file=outfile)

    def empty(self):
            return self._lossy + self._perfect == 0

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

##@package openzgy.impl.histogram

import numpy as np

class HistogramData:
    def __init__(self, range_hint=None, dtype=np.float32):
        self._hmin, self._hmax = self._suggestHistogramRange(range_hint, dtype)
        self._dtype = dtype
        self._size = 256
        # Uncomment the next lines to enable the temporary large histogram.
        # One problem is that the histogram may lose the zero centric
        # property that the user specified range might have, making the
        # result look bad.
        #if np.issubdtype(dtype, np.integer):
        #    self._size = max(256, min(65536, self._hmax - self._hmin + 1))
        self._bins = np.zeros(self._size, dtype=np.int64)
        if False:
            print("@ Histogram ", self._hmin, self._hmax, "for data", range_hint)

    @staticmethod
    def _suggestHistogramRange(range_hint, dtype):
        """
        Choose the histogram range to use.

        The histogram range is normally set to the coding range for
        integral data and the actual data range for float data.
        This method takes care of corner cases and (future) subtle
        tewaks such as making the range zero centric.
        """
        # This logic in C++ is found in GenLodImpl::suggestHistogramRange()
        # and has a much more detailed explanation of what is going on.
        bogus = (-128, +127)
        if np.issubdtype(dtype, np.integer):
            # The histogram is built from storage values so its maximum
            # possible range is already known. For int8 there will be one
            # bin for every possible value. For int16 we might do the same,
            # temporarily producing a histogram with 65,536 values and then
            # whittling it down to 256 bins before storing it. But for now
            # just map to the user provided coding range and hope that the
            # user didn't decide to use just a small part of the available
            # integer storage values.
            return (np.iinfo(dtype).min, np.iinfo(dtype).max)
        else:
            # Choose histogram range based on the min/max value collected
            # while writing lod 0. Always end up with a sane interval with
            # min < max to avoid problems generating the histogram and also
            # for applications reading the file later. For completely empty
            # files just use (-1,+1) which is as good a default as any.
            if (not range_hint or
                not np.isfinite(range_hint[0]) or
                not np.isfinite(range_hint[1]) or
                range_hint[0] > range_hint[1]):
                return bogus # nothing written or error.
            elif range_hint[0] < range_hint[1]:
                # This is the normal case for floating point data.
                # Don't return numpy types. They have weird rules.
                return (float(range_hint[0]), float(range_hint[1]))
            elif range_hint[0] > 0: # At this point, hint[0] == hint[1]
                return (0, range_hint[0]) # single positive value
            elif range_hint[0] < 0:
                return (range_hint[0], 0) # single negative value
            else:
                return bogus # all zero

    def _histogram_data(self, data):
        if not np.issubdtype(self._dtype, np.integer):
            # numpy.histogram is documented to ignore values outside range.
            # Handling of NaN is undocumented and currently reports errors
            # from low level code. So, map NaN to +Inf just in case.
            # TODO-Performance: This is not really a good idea.
            data = np.copy(data)
            data[np.isnan(data)] = np.inf
        return np.histogram(data, bins=self._size, range=self.np_range)[0]

    def add(self, data, factor = 1):
        tmp = self._histogram_data(data)
        if factor != 1:
            tmp *= factor
        self._bins += tmp

    def scale(self, a, b):
        self._hmin = a * self._hmin + b
        self._hmax = a * self._hmax + b

    def resize(self, newsize):
        binwidth = (self._hmax - self._hmin) / (self._size - 1)
        oldbins = self._bins
        oldsize = self._size
        self._size = newsize
        self._bins = np.zeros(self._size, dtype=np.int64)
        if np.count_nonzero(oldbins) == 0:
            return
        if newsize >= oldsize:
            self._bins[:oldsize] = oldbins
            self._hmax = self._hmin + binwidth * (self._size - 1)
            return
        skiplo = np.argmax(oldbins[::1] != 0)
        skiphi = np.argmax(oldbins[::-1] != 0)
        factor = max(1, (oldsize-skiplo-skiphi + (newsize-1)) // (newsize-2))
        factor = ((factor // 2) * 2) + 1 # Round up to make it odd.

        # Very minor issue: I reserve the first and last bin to hold
        # data from the misaligned part. If enerything ends up aligned
        # those two end up unused. I am absolutely sure no one will
        # notice. *except possibly* when running unit tests.

        # Adjust skiplo and skiphi upwards so that (a) neither moves
        # more than "factor", (b) neither becomes negative, (c) the
        # remaining size - skiphi - skiplo is a multiple of "factor",
        # and (d) any zero-centric property is preserved by making
        # sure the "zero" bin in the input ends up in the middle of
        # one of the output bins. The last one is where it gets really
        # tricky and TODO-High must be implemented. Or YAGNI, remove
        # the capability to resize.

        # Combine "factor" input bins into each output bin
        center_count = ((oldsize-skiphi-skiplo)//factor)*factor
        skiphi = oldsize - skiplo - center_count
        partial = np.sum(oldbins[skiplo:oldsize-skiphi].reshape(-1,factor),axis=1)
        # Mop up the ends that might have fewer than "factor" entries.
        head = np.sum(oldbins[:skiplo])
        tail = np.sum(oldbins[oldsize-skiphi:])
        self._bins[1:(center_count//factor)+1] = partial
        self._bins[0] = head
        self._bins[(center_count//factor)+1] = tail
        # The new binwidth must be binwidth*factor.
        # The new bin[1] corresponds to old bin[skiplo], so new bin[0]
        # must be new binwidth less than that.
        self._hmin = (self._hmin + binwidth * skiplo) - (binwidth*factor)
        self._hmax = self._hmin + (binwidth*factor) * (self._size-1)

    @property
    def bins(self):
        return self._bins

    @property
    def vv_range(self):
        """
        Histogram range, voxelvision and zgy style, with numbers
        representing the center value of the first and last bin.
        """
        return (self._hmin, self._hmax)

    @property
    def np_range(self):
        """
        Histogram range, numpy and salmon style, with numbers
        representing the edges of the first and last bin.
        """
        binwidth = (self._hmax - self._hmin) / (self._size - 1)
        return (self._hmin - binwidth/2, self._hmax + binwidth/2)

    def binvalue(self, bin_number):
        """
        Convert a single bin number to the center value of this bin.
        Note that in ZGY this will refer to storage values, so you
        may need to explicitly convert the result.
        """
        binwidth = (self._hmax - self._hmin) / (self._size - 1)
        return self._hmin + bin_number * binwidth

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

##@package openzgy.impl.zfp_compress

import math
import numpy as np
import time
import sys

from ..impl import enum as impl_enum
from ..exception import *
from ..impl.compress import CompressFactoryImpl, CompressPlugin, CompressStats, CompressionResult

try:
    zfpy_ok = False
    from zfpy import compress_numpy as zfpy_compress_numpy
    from zfpy import decompress_numpy as zfpy_decompress_numpy
    zfpy_ok = True
except Exception as ex:
    print("Warning: ZFP compression is not available:", str(ex), file=sys.stderr)

class ZfpCompressPlugin(CompressPlugin):
    """
    Implement ZFP compression. See the CompressPlugin base type
    for details. Most methods are static or class methods.
    The exception is __init__, __call__, and dump().
    """
    @staticmethod
    def compress(data, snr = 30, stats = None):
        if zfpy_ok:
            return ZfpCompressPlugin._zfp_compress(data, snr, stats)
        else:
            return None

    @staticmethod
    def decompress(cdata, status, shape, file_dtype, user_dtype):
        if zfpy_ok:
            return ZfpCompressPlugin._zfp_decompress(cdata, status, shape)
        else:
            return None

    @staticmethod
    def factory(snr = 30):
        return None if snr <= 0 else ZfpCompressPlugin(snr)

    def __init__(self, snr = 30):
        self._snr = snr
        self._details = "ZFP[target_snr={0:.1f}]".format(snr)
        self._stats = CompressStats(self._details)

    def __call__(self, data):
        return self.compress(data, self._snr, self._stats)

    def __str__(self):
        return self._details

    def dump(self, msg=None, *, outfile=None, text=True, csv=False, reset=True):
        if not self._stats.empty():
            self._stats.dump(msg, outfile=outfile, text=text, csv=csv)
        if reset:
            self._stats = CompressStats(self._details)

    @staticmethod
    def _zfp_precision_from_snr(data, want_snr):
        # Not sure I want this test...
        if want_snr < 10 or want_snr > 70:
            return 0, 0, 5

        if not np.issubdtype(data.dtype, np.integer):
            if not np.all(np.isfinite(data)):
                return 0, 0, 5 # Only lossless and no-compression handles this.

        # ZFP precision has a valid range of 1 to 64 bit planes.
        # If compression is done simply by reducing the number of bits:
        # If signal uses 8 bits (-128..+127) the average magnitude of the
        # signal would be 64 and the acerage quantization noise 0.5.
        # So SNR with the current metric would be 20*log10(128) ~= 42 dB.
        # Assume ZFP is better than that if asked to use N bit planes:
        # 48 dB at 8 bits, or simply 6 dB per bit.
        # A heuristic taken from hca_filt.zgy gets a somewhat different
        # result, but is quite accurate for that data set.
        # TODO-Test: test on other data sets; test on int8 and int16 as well.
        #return (want_snr + 3) // 6  if want_snr >= 10 and want_snr <= 70 else 0
        precision = (int(want_snr) + 23) // 5
        snr_rounded = (precision * 5) - 20
        return precision, snr_rounded, 5

    @classmethod
    def _zfp_try_one_lossless(cls, idata, file_dtype, snr_wanted):
        """
        Lossless compression of the input. snr_wanted is only used for
        logging since the actual snr will be perfect.
        """
        ctime = None
        dtime = None
        ddata = idata
        starttime = time.perf_counter()
        cdata = zfpy_compress_numpy(idata.astype(np.float32, copy=False))
        if True: # For testing only, for logging compression and time used.
            ctime = time.perf_counter() - starttime
            ddata = zfpy_decompress_numpy(cdata).astype(idata.dtype)
            dtime = time.perf_counter() - starttime - ctime
            if not np.allclose(idata, ddata, equal_nan=True):
                raise InternalError("zfp: claims to be lossless but isn't.")
        isize = idata.size * file_dtype.itemsize
        csize = len(cdata)
        # Noise should be 0 and snr 99, but I go thru the motions
        # to get accurate statistics. "signal" gets logged.
        if np.all(np.isfinite(idata)):
            signal, noise = CompressStats._collect_snr(idata, ddata)
        else:
            signal, noise = (0, 0)
        snr_result = CompressStats._compute_snr(signal, noise)
        return CompressionResult(cdata       = cdata,
                                 csize       = csize,
                                 isize       = isize,
                                 signal      = signal,
                                 noise       = noise,
                                 ctime       = ctime,
                                 dtime       = dtime,
                                 snr_result  = snr_result,
                                 snr_wanted  = snr_wanted,
                                 snr_rounded = 99,
                                 snr_step    = 0)

    @classmethod
    def _zfp_try_one_precision(cls, idata, file_dtype, snr_wanted):
        # the next method checks for non-finite numbers.
        precision, snr_rounded, snr_step = cls._zfp_precision_from_snr(idata, snr_wanted)
        if precision:
            starttime = time.perf_counter()
            cdata = zfpy_compress_numpy(idata, precision=precision)
            ctime = time.perf_counter() - starttime
            ddata = zfpy_decompress_numpy(cdata)
            dtime = time.perf_counter() - starttime - ctime
            isize = idata.size * file_dtype.itemsize
            csize = len(cdata)
            signal, noise = CompressStats._collect_snr(idata, ddata)
            snr_result = CompressStats._compute_snr(signal, noise)
            return CompressionResult(cdata       = cdata,
                                     csize       = csize,
                                     isize       = isize,
                                     signal      = signal,
                                     noise       = noise,
                                     ctime       = ctime,
                                     dtime       = dtime,
                                     snr_result  = snr_result,
                                     snr_wanted  = snr_wanted,
                                     snr_rounded = snr_rounded,
                                     snr_step    = snr_step)
        else:
            return CompressionResult(*((None,)*11))

    @classmethod
    def _zfp_compress_precision(cls, idata, file_dtype, want_snr, stats):
        r = cls._zfp_try_one_precision(idata, file_dtype, want_snr)
        if r.cdata is None or r.csize > 0.9*r.isize:
            r = cls._zfp_try_one_lossless(idata, file_dtype, want_snr)
        if r.cdata is not None:
            # need to make sure the compressed size is not larger than
            # the input. Note that r.isize needs to be computed as the size
            # the brick has on disk. If idata has been converted from int8
            # to int32 or float32 it is still the size on disk that matters.
            # The test below should hardly ever fail for int16 or float data
            # but for int8 it is a real possibility.
            if r.csize <= 0.9 * r.isize:
                stats.add(signal=r.signal, noise=r.noise, snr=r.snr_result,
                          isize=r.isize, csize=r.csize,
                          ctime=r.ctime, dtime=r.dtime,
                          msg="forced")
                return r.cdata
        return None

    @classmethod
    def _zfp_compress(cls, data, want_snr = 30, stats = None):
        """
        Compression plug-in offering ZFP compression of data bricks.

        The function can be passed to _writeRegion and _writeAllLODs
        as-is but will in that case have a hard coded snr of 30.
        And no compression statistics will be recorded.
        To be able to change the snr use something like
            lambda x, snr=snr: _zfp_compress(x, snr)
        which will capture your local "snr" variable.

        The input is a 3d or (TODO-Low 2d) numpy array and the output is bytes.
        ZFP or whatever algorithm is used is assumed to handle big / little
        endian conversion itself. TODO-Worry this is not quite true for ZFP.
        See the documentation. A special compilation flag is needed
        on bug endian machines. Also I suspect the optional hedaer
        (which this code uses) might need byte swapping.
        """
        # If forced inside test_compress.run_file, make sure we are really just used to handle float data.
        #assert data.dtype == np.float32

        if want_snr < 0 or not zfpy_ok:
            return None # will end up uncompressed.

        # Note, the api currently enforces file_dtype == np.float32,
        # to avoid having the user shoot himself in the foot.

        file_dtype = data.dtype
        data = data.astype(np.float32, copy=False)

        cdata = cls._zfp_compress_precision(data, file_dtype, want_snr, stats)
        if cdata is not None: return cdata

        # Give up. Do not compress this brick.
        # Not sure whether I want to count this brick in the statistics.
        # Definitely don't include ctime and dtime; timing here is irrelevant.
        if stats:
            isize = data.size * data.itemsize
            signal, noise = CompressStats._collect_snr(data, data)
            stats.add(signal=signal, noise=noise, snr=99,
                      isize=isize, csize=isize, msg="noncompr")
        return None

    @classmethod
    def _zfp_decompress(cls, cdata, status, shape):
        """
        Decompress data produced by _zfp_compress.
        ZFP will return an ndarray and it is assumed that any byte
        swapping has already been taken care of. ZFP encodes size in
        its own header so we ignore what is passed by the caller. ZFP
        also encodes dtype, but doesn't recognize int8 or int16 so
        thise will show up as float or int32 and must be converted.

        We need to be told which data type the caller eventually wants,
        because the file might contain integral data encoded as float.
        If the caller wants float data it would add more noise if we
        temporarily convert it to int here.

        See CompressPlugin.decompress for the argument list etc.
        """
        if len(cdata) < 4 or cdata[:3] != bytes("zfp", "ASCII"):
            return None # Not ours.
        # Might be e.g. a memoryview if reading from cloud.
        if not cdata is bytes: cdata = bytes(cdata)
        ddata = zfpy_decompress_numpy(cdata)
        #print("Inflated", len(cdata), "->", ddata.size*ddata.itemsize)
        return ddata

# Add to list of known factories.
if zfpy_ok:
    CompressFactoryImpl.registerCompressor("ZFP", ZfpCompressPlugin.factory)
    CompressFactoryImpl.registerDecompressor("ZFP", ZfpCompressPlugin.decompress)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
